# Module for creating Blu-ray Disc format chunked files
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package BDChunkedFile;

use strict;

###########################################################################
############################# Class interface #############################
###########################################################################

# create:  Create a new Blu-ray Disc format chunked file instance.
#
# [Parameters]
#     $type: File type (1-8 character string)
# [Return value]
#     New file instance

sub create($)
{
    my ($class, $type) = @_;

    die "No type given\n" if length($type) < 1;
    die "Type $type too long\n" if length($type) > 8;

    my $self = {};
    bless $self, $class;

    $self->{type} = $type;
    $self->{chunks} = [];

    return $self;
}

###########################################################################

# add_chunk:  Add a new chunk to a Blu-ray Disc format chunked file.
#
# [Parameters]
#     $chunk: Chunk data
# [Return value]
#     None

sub add_chunk($)
{
    my ($self, $chunk) = @_;

    if (scalar(@{$self->{chunks}}) >= 9) {
        die "$self->{type}: Too many chunks\n";
    }
    push @{$self->{chunks}}, $chunk;
}

###########################################################################

# generate:  Generate the file data for a Blu-ray Disc format chunked file.
#
# [Parameters]
#     None
# [Return value]
#     File data, as a string

sub generate()
{
    my ($self) = @_;

    my $result = substr($self->{type} . ("\0" x 8), 0, 8);

    my @offset_list = ();
    my $offset = 8 + 4*8;
    foreach my $chunk (@{$self->{chunks}}) {
        push @offset_list, $offset;
        $offset += 4 + length($chunk);
    }
    shift @offset_list;  # Discard the first offset
    $result .= pack("N8", @offset_list);

    foreach my $chunk (@{$self->{chunks}}) {
        $result .= pack("N", length($chunk));
        $result .= $chunk;
    }

    return $result;
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

# type
#     The file type (e.g. "MPLS0100").
# chunks
#     An array containing all file chunks in order.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
