# BDDisc.pm -- module for creating a Blu-ray Disc file structure
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package BDDisc;

use strict;

use BDChunkedFile;
use BDMenu;
use BDSimpleMenu;
use M2TS;

# FIXME: these ought to be declared in a common file
use constant STREAM_TYPE_MPEG2 => 0x02;
use constant STREAM_TYPE_AAC   => 0x0F;
use constant STREAM_TYPE_H264  => 0x1B;
use constant STREAM_TYPE_LPCM  => 0x80;
use constant STREAM_TYPE_AC3   => 0x81;
use constant STREAM_TYPE_SUB   => 0x90;  # Subpicture stream(?)
use constant STREAM_TYPE_MENU  => 0x91;  # Menu control stream

###########################################################################

use constant MENU_M2TS_INDEX => 1000;

# BDMV register usage:
#    Registers 0-255: temporaries
#    Register 256: current menu stream
#    Register 257: current title
#    Register 258: initial chapter when starting a title (0 to start at
#       the beginning of the title)
use constant REG_cur_menu  => 256;
use constant REG_cur_title => 257;
use constant REG_chapter   => 258;

###########################################################################
############################# Class interface #############################
###########################################################################

# create:  Create a new Blu-ray Disc file structure instance.
#
# [Parameters]
#        $out_path: Pathname of the directory under which all files will be
#                      created (the directory must already exist)
#      $AVCHD_mode: True to omit BDMV-specific directories (CERTIFICATE,
#                      etc.) from the filesystem, else false
#        $FAT_mode: True to rename files for writing to a FAT filesystem
#                      (8.3 filename format)
#      $split_size: Size in bytes at which to split M2TS files, or undef
#                      to not split
# [Return value]
#     New file structure instance

sub create($$$$)
{
    my ($class, $out_path, $AVCHD_mode, $FAT_mode, $split_size) = @_;
    die "$out_path: Not found or not a directory\n" if ! -d $out_path;

    # Create the instance itself
    my $self = {};
    bless $self, $class;

    $self->{out_path}    = $out_path;
    $self->{AVCHD_mode}  = $AVCHD_mode;
    $self->{FAT_mode}    = $FAT_mode;
    $self->{split_size}  = $split_size;
    $self->{m2ts_num}    = 1;
    $self->{clipinfo}    = [];
    $self->{has_menu}    = 0;
    $self->{menu_pages}  = 0;

    # Create the directory structure
    # (Note: the space in the prototype is to keep Emacs from getting confused)
    sub mkdir_or_die($ ) {-d $_[0] or mkdir $_[0] or die "mkdir($_[0]): $!";}
    mkdir_or_die "$out_path/BDMV";
    mkdir_or_die "$out_path/BDMV/BACKUP";
    mkdir_or_die "$out_path/BDMV/CLIPINF";
    mkdir_or_die "$out_path/BDMV/PLAYLIST";
    mkdir_or_die "$out_path/BDMV/STREAM";
    if (!$AVCHD_mode) {
        mkdir_or_die "$out_path/BDMV/AUXDATA";
        mkdir_or_die "$out_path/BDMV/BACKUP/BDJO";
        mkdir_or_die "$out_path/BDMV/BDJO";
        mkdir_or_die "$out_path/BDMV/JAR";
        mkdir_or_die "$out_path/BDMV/META";
        mkdir_or_die "$out_path/CERTIFICATE";
        mkdir_or_die "$out_path/CERTIFICATE/BACKUP";
    }

    # Success
    return $self;
}

###########################################################################

# add_stream:  Add an M2TS stream to a Blu-ray Disc file structure.
#
# [Parameters]
#                 $path: Pathname of the M2TS stream
#             $chapters: Chapters to define; one of:
#                           - undef (don't define any chapters)
#                           - Scalar value (define chapters every N seconds)
#                           - List reference (chapter positions in seconds)
#          $post_action: Action to taken when the stream ends; one of:
#                           - "standby" (go to standby mode)
#                           - "menu" (return to the menu)
#                           - "title=N[.C]" (go to title N [chapter C])
#     $exclude_pids_ref: List (reference) containing PIDs to be excluded
#                           from the M2TS stream (optional)
# [Return value]
#     None

sub add_stream($;$)
{
    my ($self, $path, $chapters, $post_action, $exclude_pids_ref) = @_;

    if ($post_action ne "standby"
     && $post_action ne "menu"
     && $post_action !~ /^title=0*[1-9][0-9]*(?:\.0*[1-9][0-9]*)?$/
    ) {
        die "Invalid post action: $post_action\n";
    }

    my $clipinfo = $self->copy_ts('source'   => $path,
                                  'destdir'  => "$self->{out_path}/BDMV/STREAM",
                                  'm2ts_num' => $self->{m2ts_num},
                                  'split'    => $self->{split_size},
                                  'chapters' => $chapters,
                                  'exclude'  => $exclude_pids_ref);
    for (my $i = 0; $i < scalar(@{$clipinfo->{outfiles}}); $i++) {
        my $dest_num = $clipinfo->{outfiles}[$i]{m2ts_num};
        $self->gen_clpi($clipinfo, $i,
                        sprintf("%s/BDMV/CLIPINF/%05d.%s",
                                $self->{out_path}, $dest_num,
                                $self->{FAT_mode} ? "CPI" : "clpi"));
    }
    $self->gen_mpls($clipinfo,
                    sprintf("%s/BDMV/PLAYLIST/%05d.%s", $self->{out_path},
                            scalar(@{$self->{clipinfo}}),
                            $self->{FAT_mode} ? "MPL" : "mpls"));

    $self->{m2ts_num} += scalar(@{$clipinfo->{outfiles}});
    $clipinfo->{post} = $post_action;
    push @{$self->{clipinfo}}, $clipinfo;
}

###########################################################################

# add_menu:  Add a menu to a Blu-ray Disc file structure.
#
# [Parameters]
#     $menu: The menu (BDMenu or BDSimpleMenu instance) to add
# [Return value]
#     None

sub add_menu($)
{
    my ($self, $menu) = @_;

    my $menu_program = scalar(@{$self->{clipinfo}}) + 1;

    $self->{m2ts_num} = MENU_M2TS_INDEX;
    for (my $page = 0; $page < $menu->num_pages; $page++) {
        # "split" is intentionally not passed to ensure that each menu page
        # only takes up one file
        my $info = $self->write_ts('data'     => $menu->generate_page($page, MENU_M2TS_INDEX, $menu_program),
                                   'destdir'  => "$self->{out_path}/BDMV/STREAM",
                                   'm2ts_num' => $self->{m2ts_num});
        die "Too many M2TS files (".scalar(@{$info->{outfiles}}).") for menu page $page\n"
            if scalar(@{$info->{outfiles}}) > 1;
        $self->gen_clpi($info, 0,
                        sprintf("%s/BDMV/CLIPINF/%05d.%s", $self->{out_path},
                                $self->{m2ts_num},
                                $self->{FAT_mode} ? "CPI" : "clpi"));
        $self->gen_mpls($info,
                        sprintf("%s/BDMV/PLAYLIST/%05d.%s", $self->{out_path},
                                $self->{m2ts_num},
                                $self->{FAT_mode} ? "MPL" : "mpls"));
        $self->{m2ts_num}++;
        push @{$self->{clipinfo}}, $info;
    }
    $self->{has_menu} = 1;
    $self->{menu_pages} = $menu->num_pages;
}

###########################################################################

# finalize:  Finalize the Blu-ray Disc file structure.
#
# [Parameters]
#     None
# [Return value]
#     None

sub finalize()
{
    my ($self) = @_;

    my $dir = "$self->{out_path}/BDMV/";
    $self->gen_index($dir . ($self->{FAT_mode} ? "INDEX.BDM" : "index.bdmv"));
    $self->gen_movieobj($dir . ($self->{FAT_mode} ? "MOVIEOBJ.BDM"
                                                  : "MovieObject.bdmv"));

    system "cp -a '$dir'/*." . ($self->{FAT_mode} ? "BDM" : "bdmv")
           . " '$dir'/CLIPINF '$dir'/PLAYLIST '$dir'/BACKUP/";
    die "Failed to create backup files" if $? != 0;
}

###########################################################################
############################# Local functions #############################
###########################################################################

# gen_clpi:  Generates a CLPI (clip info) file for the M2TS file described
# by the passed-in clip information and output file index.  Exits with
# failure on error.
#
# [Parameters]
#      $clipinfo_ref: Reference to a hash returned by copy_ts() describing
#                        the clip to generate the CLPI file for
#     $outfile_index: Index of the M2TS output file to generate the CLPI
#                        file for
#              $dest: Pathname for the output CLPI file
# [Return value]
#     None

sub gen_clpi($$$)
{
    my ($self, $clipinfo_ref, $outfile_index, $dest) = @_;
    my %clipinfo = %$clipinfo_ref;
    my %fileinfo = %{$clipinfo{outfiles}[$outfile_index]};

    my $clpi = BDChunkedFile->create("HDMV0100");

    my $clip_size = $fileinfo{pts_end_packet} * 192;
    my $clip_len = ($fileinfo{pts_end} - $fileinfo{pts_start}) / 90000;
    my $avg_rate = int($clip_size / $clip_len + 0.5);
    $avg_rate = 18000000/8 if $avg_rate > 18000000/8;  # FIXME: for menus
    # FIXME: Should this be the maximum rather than average rate?
    my $clip_info = pack("NNNNZ127NA4Z25",
                         0x101, 0, $avg_rate,
                         $fileinfo{num_packets}, "", 0x1E80, "HDMV", "");
    $clpi->add_chunk($clip_info);

    my $sequence_info = pack("nnNnNNN", 1, 0, 0x100, $clipinfo{pcr_pid},
                             0, int($fileinfo{pts_start}/2),
                             int($fileinfo{pts_end}/2));
    $clpi->add_chunk($sequence_info);

    my $program_info = pack("nNnCC", 1, 0, $clipinfo{pmt_pid},
                            scalar(@{$clipinfo{streams}}), 0);
    foreach my $pid (@{$clipinfo{streams}}) {
        my $streaminfo = $clipinfo{streaminfo}{$pid};
        my $stream_desc = pack("C", $streaminfo->{type});
        if ($streaminfo->{type} == STREAM_TYPE_H264
         || $streaminfo->{type} == STREAM_TYPE_MPEG2
        ) {
            my $format_code = ($streaminfo->{format} eq '480i'  ? 1 :
                               $streaminfo->{format} eq '576i'  ? 2 :
                               $streaminfo->{format} eq '480p'  ? 3 :
                               $streaminfo->{format} eq '1080i' ? 4 :
                               $streaminfo->{format} eq '720p'  ? 5 :
                               $streaminfo->{format} eq '1080p' ? 6 :
                               $streaminfo->{format} eq '576p'  ? 7 : 0);
            die "Invalid video format $streaminfo->{format}\n" if !$format_code;
            my $rate_code = ($streaminfo->{frame_rate} == 23 ? 1 :
                             $streaminfo->{frame_rate} == 24 ? 2 :
                             $streaminfo->{frame_rate} == 25 ? 3 :
                             $streaminfo->{frame_rate} == 29 ? 4 :
                             $streaminfo->{frame_rate} == 30 ? 5 :
                             $streaminfo->{frame_rate} == 59 ? 6 : 0);
            die "Invalid frame rate $streaminfo->{frame_rate}\n" if !$rate_code;
            my $aspect_code = ($streaminfo->{widescreen} ? 3 : 2);
            $stream_desc .= pack("CCn", $format_code<<4 | $rate_code,
                                 $aspect_code<<4, 0);
            $stream_desc .= "\0" x 16;
        } elsif ($streaminfo->{type} == STREAM_TYPE_LPCM
              || $streaminfo->{type} == STREAM_TYPE_AC3
              || $streaminfo->{type} == STREAM_TYPE_AAC
        ) {
            my $format_code = ($streaminfo->{channels} == 1 ? 1 :
                               $streaminfo->{channels} == 2 ? 3 : 6);
            my $rate_code = 1;  # 48kHz
            $stream_desc .= pack("C", $format_code<<4 | $rate_code);
            $stream_desc .= pack("A3", $streaminfo->{lang} || "und");
            $stream_desc .= "\0" x 16;
        } elsif ($streaminfo->{type} == STREAM_TYPE_SUB
              || $streaminfo->{type} == STREAM_TYPE_MENU) {
            $stream_desc .= pack("A3", $streaminfo->{lang} || "und");
            $stream_desc .= "\0" x 17;
        } else {
            die "Unknown stream type $streaminfo->{type}\n";
        }
        $program_info .= pack("nC", $pid, length($stream_desc)) . $stream_desc;
    }
    $clpi->add_chunk($program_info);

    my $seek_table = "";
    my $keyframe_table = "";
    my $last_pts_high = -1;
    my $last_packet_high = -1;
    for (my $keyframe = 0; $keyframe < scalar(@{$fileinfo{keyframes}});
         $keyframe++
    ) {
        my $keyframe_info = $fileinfo{keyframes}[$keyframe];
        my $pts_512 = int($keyframe_info->{pts} / 512);
        my $packet_num = $keyframe_info->{packet_num};
        my $pts_high = $pts_512 >> 11;
        my $pts_low = $pts_512 & 0x7FF;
        my $packet_high = $packet_num >> 17;
        my $packet_low = $packet_num & 0x1FFFF;
        if ($pts_high != $last_pts_high
         || $packet_high != $last_packet_high
        ) {
            $seek_table .= pack("NN",
                                $keyframe<<14 | $pts_high<<1 | $pts_low>>10,
                                $packet_num);
            $last_pts_high = $pts_high;
            $last_packet_high = $packet_high;
        }
        my $frame_size = $keyframe_info->{size} * 192;
        my $size_idx;
        if ($frame_size <= 0x20000) {
            $size_idx = 1;
        } elsif ($frame_size <= 0x40000) {
            $size_idx = 2;
        } elsif ($frame_size <= 0x60000) {
            $size_idx = 3;
        } elsif ($frame_size <= 0x90000) {
            $size_idx = 4;
        } elsif ($frame_size <= 0xE0000) {  # FIXME: Might be inaccurate
            $size_idx = 5;
        } else {  # FIXME: Not sure where 7 starts
            $size_idx = 6;
        }
        $keyframe_table .= pack("N", $size_idx<<28 | $pts_low<<17 | $packet_low);
    }
    my $seek_info = pack("nnnnnnNN", 1, 1, $clipinfo{video_pid}, 4,
                         int(length($seek_table)/2),
                         int(length($keyframe_table)/4), 0xE,
                         4 + length($seek_table));
    $seek_info .= $seek_table;
    $seek_info .= $keyframe_table;
    $clpi->add_chunk($seek_info);

    $clpi->add_chunk("");

    my $clex = BDChunkedFile->create("CLEX");
    $clex->add_chunk("\1\3\3\0");
    $clex->add_chunk(pack("nnnnnnNn", 1, 0x100, $clipinfo{video_pid},
                          0xB1B, 0x264, 0x28, 0, 0));
    $clex = $clex->generate();
    $clex = pack("NNNNN", 0x18, 1, 0x10000100, 0x18, length($clex)) . $clex;
    $clpi->add_chunk($clex);

    local *OUT;
    open OUT, ">$dest" or die "$dest: $!\n";
    binmode OUT;
    print OUT $clpi->generate();
    close OUT;
}

###########################################################################

# gen_mpls:  Generates an MPLS (play list) file for the stream described by
# the passed-in hash reference (a hash returned by &copy_ts).  Exits with
# failure on error.
#
# [Parameters]
#     $clipinfo_ref: Reference to a hash returned by copy_ts() describing
#                       the clip to generate the MPLS file for
#             $dest: Pathname for the output MPLS file
# [Return value]
#     None

sub gen_mpls($$)
{
    my ($self, $clipinfo_ref, $dest) = @_;
    my %clipinfo = %$clipinfo_ref;

    my $num_files = scalar(@{$clipinfo{outfiles}});
    my ($num_video, $num_audio, $num_subtitle, $num_menu);
    foreach my $pid (@{$clipinfo{streams}}) {
        my $streaminfo = $clipinfo{streaminfo}{$pid};
        if ($streaminfo->{type} == STREAM_TYPE_H264
         || $streaminfo->{type} == STREAM_TYPE_MPEG2
        ) {
            $num_video++;
        } elsif ($streaminfo->{type} == STREAM_TYPE_LPCM
              || $streaminfo->{type} == STREAM_TYPE_AC3
              || $streaminfo->{type} == STREAM_TYPE_AAC
        ) {
            $num_audio++;
        } elsif ($streaminfo->{type} == STREAM_TYPE_SUB) {
            $num_subtitle++;
        } elsif ($streaminfo->{type} == STREAM_TYPE_MENU) {
            $num_menu++;
        } else {
            die "Unknown stream type $streaminfo->{type}\n";
        }
    }

    my $mpls = BDChunkedFile->create("MPLS0100");

    $mpls->add_chunk(pack("nNNN", 1, $num_menu ? 0x3CB8 : 0, 0,
                          $num_menu ? 0x8000 : 0));

    my $chunk1 = pack("Nn", $num_files, 0);
    for (my $i = 0; $i < $num_files; $i++) {
        my $subinfo = pack("nCCCCNN", 0,
                           $num_video, $num_audio, $num_subtitle, $num_menu,
                           0, 0);
        foreach my $pid (@{$clipinfo{streams}}) {
            $subinfo .= pack("nnnnn", 0x901, $pid, 0, 0, 0);
            my $streaminfo = $clipinfo{streaminfo}{$pid};
            if ($streaminfo->{type} == STREAM_TYPE_H264
             || $streaminfo->{type} == STREAM_TYPE_MPEG2
            ) {
                my $format_code = ($streaminfo->{format} eq '480i'  ? 1 :
                                   $streaminfo->{format} eq '576i'  ? 2 :
                                   $streaminfo->{format} eq '480p'  ? 3 :
                                   $streaminfo->{format} eq '1080i' ? 4 :
                                   $streaminfo->{format} eq '720p'  ? 5 :
                                   $streaminfo->{format} eq '1080p' ? 6 :
                                   $streaminfo->{format} eq '576p'  ? 7 : 0);
                die "Invalid video format $streaminfo->{format}\n"
                    if !$format_code;
                my $rate_code = ($streaminfo->{frame_rate} == 23 ? 1 :
                                 $streaminfo->{frame_rate} == 24 ? 2 :
                                 $streaminfo->{frame_rate} == 25 ? 3 :
                                 $streaminfo->{frame_rate} == 29 ? 4 :
                                 $streaminfo->{frame_rate} == 30 ? 5 :
                                 $streaminfo->{frame_rate} == 59 ? 6 : 0);
                die "Invalid frame rate $streaminfo->{frame_rate}\n"
                    if !$rate_code;
                $subinfo .= pack("CCCCCC", 5,
                                 $streaminfo->{type},
                                 $format_code<<4 | $rate_code,
                                 0, 0, 0);
            } elsif ($streaminfo->{type} == STREAM_TYPE_LPCM
                  || $streaminfo->{type} == STREAM_TYPE_AC3
                  || $streaminfo->{type} == STREAM_TYPE_AAC
            ) {
                my $format_code = ($streaminfo->{channels} == 1 ? 1 :
                                   $streaminfo->{channels} == 2 ? 3 : 6);
                my $rate_code = 1;  # 48kHz
                $subinfo .= pack("CCCA3", 5,
                                 $streaminfo->{type},
                                 $format_code<<4 | $rate_code,
                                 $streaminfo->{lang} || "und");
            } elsif ($streaminfo->{type} == STREAM_TYPE_SUB
                  || $streaminfo->{type} == STREAM_TYPE_MENU
            ) {
                $subinfo .= pack("CCA3C", 5,
                                 $streaminfo->{type},
                                 $streaminfo->{lang} || "und",
                                 0);
            } else {
                die "impossible";
            }
        }
        my $fileinfo = $clipinfo{outfiles}[$i];
        my $filedata = sprintf("%05dM2TS", $fileinfo->{m2ts_num}) . "\0"
                     . pack("nNNNNnnn", $i==0 ? 0x100 : 0x600,
                            int($fileinfo->{pts_start}/2),
                            int($fileinfo->{pts_end}/2), 0, 0,
                            $num_menu ? 2 : 0, 0,
                            length($subinfo))
                     . $subinfo;
        $chunk1 .= pack("n", length($filedata)) . $filedata;
    }
    $mpls->add_chunk($chunk1);

    $clipinfo{chapters} ||= [$clipinfo{outfiles}[0]{pts_start}];
    my $num_chapters = @{$clipinfo{chapters}};
    my $chunk2 = pack("n", $num_chapters);
    for (my $i = 0; $i < $num_chapters; $i++) {
        my $file_index = 0;
        while ($file_index+1 < scalar(@{$clipinfo{outfiles}})
               && $clipinfo{chapters}[$i] >= $clipinfo{outfiles}[$file_index+1]{pts_start}
        ) {
            $file_index++;
        }
        $chunk2 .= pack("nnNnN", 1, $file_index,
                        int($clipinfo{chapters}[$i]/2), 0xFFFF, 0);
    }
    $mpls->add_chunk($chunk2);

    my $plex = BDChunkedFile->create("PLEX");
    my $title = $clipinfo{title};
    $title = pack("CZ255", length($title), $title);  #FIXME: will truncate a 255-byte-long entry
    my ($sec,$min,$hour,$day,$month,$year) = localtime($clipinfo{timestamp});
    $year += 1900;
    $month++;
    sub bcd($ ) {return int($_[0]/10)<<4 | $_[0]%10;}
    sub bcd4($ ) {return bcd(int($_[0]/100))<<8 | bcd($_[0]%100);}
    $title = pack("NNnCnCCCCCCC", 0x01030300, 0, 0xFFFF, 0x1E,
                  bcd4($year), bcd($month), bcd($day),
                  bcd($hour), bcd($min), bcd($sec), 0, 0x90) . $title;
    $title .= pack("N", 0);
    $plex->add_chunk($title);
    my $extra = "bd-gen";
    $extra = pack("CZ40", length($extra), $extra);  #FIXME: will truncate a 40-byte-long entry
    $extra = pack("nNNnCnCCCCCC", 1, 0x01030300, 0, 0, 0x1E,
                  bcd4($year), bcd($month), bcd($day),
                  bcd($hour), bcd($min), bcd($sec), 0x90) . $extra;
    $extra .= pack("Nn", -1, 0);
    $plex->add_chunk($extra);
    $plex = $plex->generate();
    $plex = pack("NNNNN", 0x18, 1, 0x10000100, 0x18, length($plex)) . $plex;
    $mpls->add_chunk($plex);

    local *OUT;
    open OUT, ">$dest" or die "$dest: $!\n";
    binmode OUT;
    print OUT $mpls->generate();
    close OUT;
}

###########################################################################

# gen_index:  Generates an index.bdmv file for the passed-in array of hash
# references (hashes returned by &copy_ts).  Exits with failure on error.
#
# [Parameters]
#     $dest: Pathname for the output index.bdmv file
# [Return value]
#     None

sub gen_index($)
{
    my ($self, $dest) = @_;
    my @cliplist = @{$self->{clipinfo}};

    my $index = BDChunkedFile->create("INDX0100");

    # Apparently the top bit here differentiates between true BDMV and AVCHD
    $index->add_chunk(pack("NNNNNNNNn", $self->{AVCHD_mode} ? 0x80000000 : 0));

    my $chunk1;
    if ($self->{AVCHD_mode}) {
        $chunk1 = pack("NNNNNNn",
                       0x40000000, 0x40000000, 0, 0x40000000, 0x4000FFFF, 0,
                       scalar(@cliplist));
    } else {
        $chunk1 = pack("NNNNNNn",
                       0x40000000, 0x40000000, 0,  # 2nd word is a program number?
                       0x40000000, 0x40000000, 0,  # 2nd word is first program to execute on disc start?
                       scalar(@cliplist));
    }
    for (my $i = 0; $i < @cliplist; $i++) {
        $chunk1 .= pack("NNN", 0x40000000, $i+1, 0);
    }
    $index->add_chunk($chunk1);

    my $idex = BDChunkedFile->create("IDEX");
    my $title = "";
    $title = pack("CZ255", length($title), $title);  #FIXME: will truncate a 255-byte-long entry
    my ($sec,$min,$hour,$day,$month,$year) = localtime(time());
    $year += 1900;
    $month++;
    sub bcd($ ) {return int($_[0]/10)<<4 | $_[0]%10;}
    sub bcd4($ ) {return bcd(int($_[0]/100))<<8 | bcd($_[0]%100);}
    $title = pack("NNNNNNNNNnnCnCCCCCCC", 0x01030300, 0, 0, 0, 0, 0, 0, 0, 0,
                  0xA, 0xFFFF, 0x1E,
                  bcd4($year), bcd($month), bcd($day),
                  bcd($hour), bcd($min), bcd($sec), 0, 0x90) . $title;
    $title .= pack("N", 0);
    $idex->add_chunk($title);
    my $extra = pack("NnNN", 2, 0, 2, scalar(@cliplist));
    for (my $i = 0; $i < @cliplist; $i++) {
        $extra .= sprintf("%05d", $i) . pack("CN", 1, $i);
    }
    $idex->add_chunk($extra);
    $idex = $idex->generate();
    $idex = pack("NNNNN", 0x18, 1, 0x10000100, 0x18, length($idex)) . $idex;
    $index->add_chunk($idex);

    local *OUT;
    open OUT, ">$dest" or die "$dest: $!\n";
    binmode OUT;
    print OUT $index->generate();
    close OUT;
}

###########################################################################

# gen_movieobj:  Generates a MovieObject.bdmv file for the passed-in array
# of hash references (hashes returned by &copy_ts).  Exits with failure on
# error.
#
# [Parameters]
#     $dest: Pathname for the output MovieObject.bdmv file
# [Return value]
#     None

sub gen_movieobj($$)
{
    my ($self, $dest) = @_;
    my @cliplist = @{$self->{clipinfo}};

    my $movieobj = BDChunkedFile->create("MOBJ0100");

    my @entries = ();

    # If there's a menu, we put it after the programs for each title;
    # note that the menu streams have already been added to @cliplist
    my $menu_program = (scalar(@cliplist) - $self->{menu_pages}) + 1;

    # Program 0 (initialization)
    push @entries, [
        # Initialize variables
        [0x50400001, REG_cur_menu, MENU_M2TS_INDEX],
        [0x50400001, REG_cur_title, 0],
        [0x50400001, REG_chapter, 0],
        # Jump to either the first title or the menu
        [0x21810000, $self->{has_menu} ? $menu_program : 1, 0],
    ];

    # Programs 1-N (one per title)
    for (my $i = 1; $i <= scalar(@cliplist) - $self->{menu_pages}; $i++) {
        my @post;
        if ($cliplist[$i-1]{post} eq "standby") {
            @post = ([0x00020000, 0, 0]);
        } elsif ($cliplist[$i-1]{post} eq "menu") {
            if (!$self->{has_menu}) {
                die "Cannot jump to menu after title $i because there is no menu\n";
            }
            @post = ([0x21810000, $menu_program, 0]);
        } elsif ($cliplist[$i-1]{post} =~ /^title=([0-9]*)(?:\.([0-9]*))?$/) {
            my ($new_title, $new_chapter) = ($1, $2);
            @post = ([0x50400001, REG_chapter, $new_chapter],
                     [0x21810000, $new_title, 0]);
        } else {
            die "impossible: bad post action $cliplist[$i-1]{post} for title $i";
        }
        push @entries, [
            # Save the current playlist number
            [0x50400001, REG_cur_title, $i-1],
            # Play the title.  Command 0x4202 takes the marker index (0...)
            # rather than the chapter number (1...), so adjust the register
            # value accordingly.
            [0x50000001, 1, REG_chapter],
            [0x50400004, 1, 1],
            [0x48400200, REG_chapter, 0],  # if chapter == 0
            [0x22000000, REG_cur_title, 0],
            [0x48400300, REG_chapter, 0],  # if chapter != 0
            [0x42020000, REG_cur_title, 1],
            # Perform appropriate post-title actions
            @post,
        ];
    }

    # Program N+1 (menu)
    push @entries, [
        # Clear the last-pressed button ID (so if we got here by an
        # auto-activate button, we don't try to activate it again)
        [0x51400003, 0x80000002, 0],
        # Play the menu stream and wait for something to happen
        [0x22000000, REG_cur_menu, 0],
    ];

    my $data = pack("Nn", 0, scalar(@entries));
    foreach my $entry (@entries) {
        $data .= pack("nn", 0x8000, scalar(@$entry));
        foreach my $command (@$entry) {
            $data .= pack("NNN", @$command);
        }
    }
    $movieobj->add_chunk($data);

    local *OUT;
    open OUT, ">$dest" or die "$dest: $!\n";
    binmode OUT;
    print OUT $movieobj->generate();
    close OUT;
}

###########################################################################
############################### TS copying ################################
###########################################################################

# copy_ts:  Copy an M2TS-format stream, returning a hash with information
# about the stream.  Exits with failure on error.
#
# Parameters are passed as a hash.  The following parameters can be set:
#     'source'   => (mandatory) The source file pathname.
#     'destdir'  => (mandatory) The directory into which output files are
#                      written.
#     'm2ts_num' => (mandatory) The number of the first M2TS to be written,
#                      e.g. 123 for "00123.m2ts".
#     'split'    => (optional)  The maximum number of bytes to write to a
#                      single M2TS file.
#     'chapters' => (optional)  Either a scalar indicating the number of
#                      seconds at which automatic chapters should be
#                      created, or a list reference containing individual
#                      chapter offsets in seconds (an initial offset of
#                      zero should _not_ be included).
#     'exclude'  => (optional)  A list reference containg PIDs to be
#                      excluded when copying the stream.
#
# The return value is a hash containing the following data:
#     'filename'    => The source filename (with directory names stripped).
#     'title'       => The stream title.  Currently, this is the same as
#                         the filename with any .m2ts or .mts extension
#                         (case insensitive) stripped.
#     'timestamp'   => The timestamp of the source file.
#     'streams'     => An array of transport stream PIDs for the data
#                         streams found in the file, in the order they were
#                         found.
#     'streaminfo'  => A hash, indexed by PID, of information about each
#                         stream (see below).
#     'chapters'    => A list of PTS values for each chapter.
#     'pmt_pid'     => The PID of the PMT stream.
#     'pcr_pid'     => The PID of the PCR stream.
#     'video_pid'   => The PID of the video stream.
#     'audio_pid'   => The PID of the audio stream.
#     'outfiles'    => An array of information about each M2TS file created
#                         (see below).
#
# The information stored in $clipinfo->{streaminfo}{$pid} for each stream
# is a hash containing the following data:
#     'type'       => The TS type of the stream (0x1B = H.264, etc.)
#     'lang'       => The 3-character language code of the stream, or undef
#                        if unknown.
#     'format'     => (video only) The video format: '480i', '720p', etc.
#     'frame_rate' => (video only) The frame rate (frames/sec), truncated
#                        to an integer.
#     'widescreen' => (video only) True if a widescreen video, else false.
#     'rate'       => (audio only) The sampling rate (samples/sec).
#     'channels'   => (audio only) The number of channels in the stream.
#
# The information stored in $clipinfo->{outfiles}[$n] for each M2TS file
# created is a hash containing the following data:
#     'm2ts_num'       => The number of the M2TS file.
#     'num_packets'    => The number of packets in the TS file.
#     'pts_start'      => The first PTS in the video stream.
#     'pts_end'        => The last PTS in the video stream.
#     'pts_end_packet' => The packet index of the 'pts_end' packet.
#     'keyframes'      => An array of keyframes, in time order.  Each entry
#                            is a hash containing two keys: 'pts' holds the
#                            PTS of the keyframe, and 'packet_num' holds
#                            the packet index (starting from 0) within the
#                            M2TS file.

sub copy_ts($)
{
    my ($self, %params) = @_;

    my $source        = $params{source};
    my $destdir       = $params{destdir};
    $self->{m2ts_num} = $params{m2ts_num};
    my $split         = $params{split};
    my $chapters      = $params{chapters};
    my $exclude       = $params{exclude};

    local *IN;
    open IN, "<$source" or die "$source: $!\n";
    binmode IN;

    my $size_packets;  # Input size in packets, undef if unknown
    if (seek(IN, 0, 2)) {
        $size_packets = int(tell(IN) / 192);
        if ($size_packets == 0) {
            die "$source is empty!\n";
        }
        seek IN, 0, 0;
    }

    my $dest = sprintf("%s/%05d.%s", $destdir, $self->{m2ts_num},
                       $self->{FAT_mode} ? "MTS" : "m2ts");
    my $m2ts = M2TS->create($dest);
    if (defined($exclude)) {
        foreach my $pid (@$exclude) {
            $m2ts->exclude_pid($pid);
        }
    }

    my %clipinfo = (filename => $source);
    $clipinfo{filename} =~ s,^.*/,,;
    $clipinfo{title} = $clipinfo{filename};
    $clipinfo{title} =~ s/\.m2?ts$//i;
    $clipinfo{timestamp} = (stat($source))[9];

    use Encode;
    my $source_display = $source;
    # FIXME: assuming a UTF-8 environment here
    Encode::_utf8_on($source_display);
    if (defined($size_packets)) {
        print STDERR "Processing $source_display...  0%\b\b\b\b";
    } else {
        print STDERR "Processing $source_display... ";
    }

    my $packets_out = 0;
    my $packet;
    while (read(IN, $packet, 192) == 192) {
        die "$source: not M2TS format or corrupt\n"
            if substr($packet,4,1) ne "G";
        if (defined($split) && ($m2ts->num_packets + 1) * 192 > $split) {
            my $old_m2ts = $m2ts;
            my $old_m2ts_num = $self->{m2ts_num};
            $self->{m2ts_num}++;
            $dest = sprintf("%s/%05d.%s", $destdir, $self->{m2ts_num},
                            $self->{FAT_mode} ? "MTS" : "m2ts");
            $m2ts = $old_m2ts->split($dest);
            push @{$clipinfo{outfiles}}, {
                m2ts_num       => $old_m2ts_num,
                num_packets    => $old_m2ts->num_packets,
                pts_start      => $old_m2ts->pts_start,
                pts_end        => $old_m2ts->pts_end,
                pts_end_packet => $old_m2ts->pts_end_packet,
                keyframes      => $old_m2ts->keyframes,
            };
        }
        $m2ts->write_packet($packet);
        $packets_out++;
        if (defined($size_packets) && $packets_out < $size_packets) {
            my $last_pct = int((($packets_out - 1) / $size_packets) * 100);
            my $pct = int(($packets_out / $size_packets) * 100);
            if ($pct != $last_pct) {
                printf STDERR " %2d%%\b\b\b\b", $pct;
            }
        }
    }

    print STDERR " done.\n";

    $m2ts->close();
    push @{$clipinfo{outfiles}}, {
        m2ts_num       => $self->{m2ts_num},
        num_packets    => $m2ts->num_packets,
        pts_start      => $m2ts->pts_start,
        pts_end        => $m2ts->pts_end,
        pts_end_packet => $m2ts->pts_end_packet,
        keyframes      => $m2ts->keyframes,
    };

    $clipinfo{streams}    = $m2ts->streams;
    $clipinfo{streaminfo} = $m2ts->streaminfo;
    $clipinfo{pmt_pid}    = $m2ts->pmt_pid;
    $clipinfo{pcr_pid}    = $m2ts->pcr_pid;
    $clipinfo{video_pid}  = $m2ts->video_pid;
    $clipinfo{audio_pid}  = $m2ts->audio_pid;

    my $pts_start = $clipinfo{outfiles}[0]{pts_start};
    my $pts_end = $m2ts->pts_end;
    my @chapter_pts = ($pts_start);
    if (defined($chapters)) {
        if (ref($chapters) eq "ARRAY") {
            foreach my $time (@$chapters) {
                my $pts = $pts_start + int($time * 90000);
                last if $pts >= $pts_end;
                push @chapter_pts, $pts;
            }
        } elsif ($chapters > 0) {
            for (my $count = 1;
                 $pts_start + int($count * $chapters * 90000) < $pts_end;
                 $count++
            ) {
                push @chapter_pts,
                    $pts_start + int($count * $chapters * 90000);
            }
        }
    }
    $clipinfo{chapters}   = [@chapter_pts];

    close IN;
    return {%clipinfo};
}

###########################################################################

# write_ts:  Write an M2TS-format stream from a memory buffer, returning a
# hash with information about the stream.  Exits with failure on error.
#
# Parameters are passed as a hash.  The following parameters can be set:
#     'data'     => (mandatory) The source data.
#     'destdir'  => (mandatory) The directory into which output files are
#                      written.
#     'm2ts_num' => (mandatory) The number of the first M2TS to be written,
#                      e.g. 123 for "00123.m2ts".
#     'split'    => (optional)  The maximum number of bytes to write to a
#                      single M2TS file.
#
# The return value is a hash in the same format as that returned by
# copy_ts().

sub write_ts
{
    my ($self, %params) = @_;

    my $data          = $params{data};
    my $destdir       = $params{destdir};
    $self->{m2ts_num} = $params{m2ts_num};
    my $split         = $params{split};

    my $dest = sprintf("%s/%05d.%s", $destdir, $self->{m2ts_num},
                       $self->{FAT_mode} ? "MTS" : "m2ts");
    my $m2ts = M2TS->create($dest);

    my %clipinfo = ();
    $clipinfo{timestamp} = time();

    for (my $index = 0; $index+192 <= length($data); $index += 192) {
        my $packet = substr($data, $index, 192);
        die "Data not M2TS format or corrupt\n" if substr($packet,4,1) ne "G";
        if (defined($split) && ($m2ts->num_packets + 1) * 192 > $split) {
            my $old_m2ts = $m2ts;
            my $old_m2ts_num = $self->{m2ts_num};
            $self->{m2ts_num}++;
            $dest = sprintf("%s/%05d.%s", $destdir, $self->{m2ts_num},
                            $self->{FAT_mode} ? "MTS" : "m2ts");
            $m2ts = $old_m2ts->split($dest);
            push @{$clipinfo{outfiles}}, {
                m2ts_num       => $old_m2ts_num,
                num_packets    => $old_m2ts->num_packets,
                pts_start      => $old_m2ts->pts_start,
                pts_end        => $old_m2ts->pts_end,
                pts_end_packet => $old_m2ts->pts_end_packet,
                keyframes      => $old_m2ts->keyframes,
            };
        }
        $m2ts->write_packet($packet);
    }
    $m2ts->close();

    push @{$clipinfo{outfiles}}, {
        m2ts_num       => $self->{m2ts_num},
        num_packets    => $m2ts->num_packets,
        pts_start      => $m2ts->pts_start,
        pts_end        => $m2ts->pts_end,
        pts_end_packet => $m2ts->pts_end_packet,
        keyframes      => $m2ts->keyframes,
    };
    $clipinfo{streams}    = $m2ts->streams;
    $clipinfo{streaminfo} = $m2ts->streaminfo;
    $clipinfo{pcr_pid}    = $m2ts->pcr_pid;
    $clipinfo{video_pid}  = $m2ts->video_pid;
    $clipinfo{audio_pid}  = $m2ts->audio_pid;

    # FIXME: quick hack for menus
    if ($clipinfo{outfiles}[0]{pts_end} == $clipinfo{outfiles}[0]{pts_start}) {
        $clipinfo{outfiles}[0]{pts_end} = $clipinfo{outfiles}[0]{pts_start} + 3003;
        $clipinfo{outfiles}[0]{pts_end_packet} = $clipinfo{outfiles}[0]{num_packets};
    }

    return {%clipinfo};
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

# out_path
#     The pathname of the directory under which all files will be created.
# AVCHD_mode
#     True if BDMV-specific directories (CERTIFICATE, etc.) should be
#     omitted.
# FAT_mode
#     True if files should be renamed for writing to a FAT filesystem (8.3
#     filename format).
# split_size
#     Size in bytes at which to split M2TS files, or undef to not split.
# m2ts_num
#     The next index to use for writing an M2TS file.
# clipinfo
#     An array of information about each clip on the disc.
# has_menu
#     True if the file structure contains a menu.
# menu_pages
#     Number of pages (i.e. streams) in the menu.  Zero if the file
#     structure does not contain a menu.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
