# Module for creating a Blu-ray Disc menu
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package BDMenu;

use strict;
use M2TSGenerator;
use PNGReader;

###########################################################################

use constant PRIVATE_STREAM_PALETTE    => 0x14;  # Color palette data
use constant PRIVATE_STREAM_GRAPHICS   => 0x15;  # Menu graphic data
use constant PRIVATE_STREAM_CONTROL    => 0x18;  # Menu control data
use constant PRIVATE_STREAM_TERMINATOR => 0x80;  # I'll Be Bach

# These are from BDDisc.pm (FIXME: code duplication)
use constant REG_cur_menu  => 256;
use constant REG_cur_title => 257;
use constant REG_chapter   => 258;

###########################################################################
############################# Class interface #############################
###########################################################################

# create:  Create a new menu instance.
#
# [Parameters]
#     None
# [Return value]
#     New menu instance

sub create()
{
    my ($class) = @_;

    my $self = {};
    bless $self, $class;

    $self->{pages}  = [];
    $self->{images} = [];

    return $self;
}

###########################################################################

# png_to_image:  Convert a PNG image for use in a menu.
#
# [Parameters]
#     $png: PNGReader object
# [Return value]
#     Image index for use as link button values

sub png_to_image($)
{
    my ($self, $png) = @_;
    use integer;

    my $image = {
        width  => $png->width(),
        height => $png->height(),
        data   => $png->data(),
    };
    foreach my $row (@{$image->{data}}) {
        foreach my $pixel (@$row) {
            my $a = $pixel>>24 & 0xFF;
            if ($a == 0) {
                $pixel = 0;
                next;
            }
            my $r = $pixel>>16 & 0xFF;
            my $g = $pixel>> 8 & 0xFF;
            my $b = $pixel>> 0 & 0xFF;
            my $Y = ((16829*$r + 33039*$g +  6416*$b + 32768) >> 16) +  16;
            my $U = ((-9714*$r - 19070*$g + 28784*$b + 32768) >> 16) + 128;
            my $V = ((28784*$r - 24103*$g -  4681*$b + 32768) >> 16) + 128;
            $pixel = $Y<<24 | $V<<16 | $U<<8 | $a;
        }
    }

    my $index = scalar(@{$self->{images}});
    push @{$self->{images}}, $image;
    return $index;
}

###########################################################################

# set_background:  Set the background parameters for a page of a menu.
#
# [Parameters]
#       $page: Page number (0...)
#      $color: Background color (0xRRGGBB)
#      $image: Background image pathname, or undef for none
#     $resize: Resize operation: "(linear|cubic|sinc)[,aspect]" or undef
# [Return value]
#     None

sub set_background($$$$)
{
    my ($self, $page, $color, $image, $resize) = @_;

    my ($resize_type, $resize_keep_aspect);
    if (defined($resize)) {
        if ($resize !~ /^\s*(linear|cubic|sinc)(\s*,\s*aspect)?\s*$/) {
            die "Invalid resize operation: $resize\n";
        }
        $resize_type = $1;
        $resize_keep_aspect = ($2 ne "");
    }

    if (!defined($self->{pages}[$page])) {
        $self->{pages}[$page] = {text => [], links => []};
    }
    $self->{pages}[$page]{background} =
        [$color, $image, $resize_type, $resize_keep_aspect];
}

###########################################################################

# add_text:  Add a text string to a page of a menu.
#
# [Parameters]
#      $page: Page number (0...)
#         $x: Base horizontal coordinate of text (0 <= $x < 1920)
#         $y: Baseline vertical coordinate of text (0 <= $y < 1080)
#     $align: Text alignment relative to $x: -1 = left, 0 = center, +1 = right
#      $font: Font name
#      $size: Font size (height in pixels)
#     $color: Font color (0xAARRGGBB)
#      $text: Text to add
# [Return value]
#     None

sub add_text($$$$$$$$)
{
    my ($self, $page, $x, $y, $align, $font, $size, $color, $text) = @_;

    if (!defined($self->{pages}[$page])) {
        $self->{pages}[$page] = {text => [], links => []};
    }
    push @{$self->{pages}[$page]{text}},
        [$x, $y, $align, $font, $size, $color, $text];
}

###########################################################################

# add_link_to_title:  Add a title link to a page of a menu.
#
# [Parameters]
#        $page: Page number (0...)
#       $index: Link index (0...65535)
#       $x, $y: Coordinates of link image
#      $normal: Link image when not highlighted (undef for none)
#      $hilite: Link image when highlighted (undef for none)
#      $active: Link image after being activated (undef for none)
#          $up: Next link to select when "up" is pressed (undef for none)
#        $down: Next link to select when "down" is pressed (undef for none)
#        $left: Next link to select when "left" is pressed (undef for none)
#       $right: Next link to select when "right" is pressed (undef for none)
#        $auto: True to make link auto-activate when highlighted
#      $target: Link target (title number, 1...)
#     $chapter: (Optional) Target chapter within title
# [Return value]
#     None

sub add_link_to_title($$$$$$$$$$$$$;$)
{
    my ($self, $page, $index, $x, $y, $normal, $hilite, $active,
        $up, $down, $left, $right, $auto, $target, $chapter) = @_;

    $normal = -1 if !defined($normal);
    $hilite = -1 if !defined($hilite);
    $active = -1 if !defined($active);
    $up     = -1 if !defined($up);
    $down   = -1 if !defined($down);
    $left   = -1 if !defined($left);
    $right  = -1 if !defined($right);

    if (!defined($self->{pages}[$page])) {
        $self->{pages}[$page] = {text => [], links => []};
    }
    $self->{pages}[$page]{links}[$index] =
        [$x, $y, $normal, $hilite, $active,
         $up, $down, $left, $right, $auto, 'TITLE', $target, $chapter];
}

###########################################################################

# add_link_to_page:  Add a page-jump link to a page of a menu.
#
# [Parameters]
#       $page: Page number (0...)
#      $index: Link index (0...65535)
#      $x, $y: Coordinates of link image
#     $normal: Link image when not highlighted (-1 for none)
#     $hilite: Link image when highlighted (-1 for none)
#     $active: Link image after being activated (-1 for none)
#         $up: Next link to select when "up" is pressed (-1 for none)
#       $down: Next link to select when "down" is pressed (-1 for none)
#       $left: Next link to select when "left" is pressed (-1 for none)
#      $right: Next link to select when "right" is pressed (-1 for none)
#       $auto: True to make link auto-activate when highlighted
#     $target: Link target (page number)
# [Return value]
#     None

sub add_link_to_page($$$$$$$$$$$$$)
{
    my ($self, $page, $index, $x, $y, $normal, $hilite, $active,
        $up, $down, $left, $right, $auto, $target) = @_;

    $normal = -1 if !defined($normal);
    $hilite = -1 if !defined($hilite);
    $active = -1 if !defined($active);
    $up     = -1 if !defined($up);
    $down   = -1 if !defined($down);
    $left   = -1 if !defined($left);
    $right  = -1 if !defined($right);

    if (!defined($self->{pages}[$page])) {
        $self->{pages}[$page] = {text => [], links => []};
    }
    $self->{pages}[$page]{links}[$index] =
        [$x, $y, $normal, $hilite, $active,
         $up, $down, $left, $right, $auto, 'PAGE', $target];
}

###########################################################################

# set_link_movement:  Set the movement targets for a link.  Any of $up,
# $down, $left, and $right can be set to undef to leave that field alone.
#
# [Parameters]
#       $page: Page number (0...)
#      $index: Link index (0...65535)
#         $up: Next link to select when "up" is pressed (-1 for none)
#       $down: Next link to select when "down" is pressed (-1 for none)
#       $left: Next link to select when "left" is pressed (-1 for none)
#      $right: Next link to select when "right" is pressed (-1 for none)
# [Return value]
#     None

sub set_link_movement($$$$$$)
{
    my ($self, $page, $index, $up, $down, $left, $right) = @_;

    if (!defined($self->{pages}[$page])) {
        $self->{pages}[$page] = {text => [], links => []};
    }
    $self->{pages}[$page]{links}[$index][5] = $up    if defined($up);
    $self->{pages}[$page]{links}[$index][6] = $down  if defined($down);
    $self->{pages}[$page]{links}[$index][7] = $left  if defined($left);
    $self->{pages}[$page]{links}[$index][8] = $right if defined($right);
}

###########################################################################

# num_pages:  Return the number of pages in the menu.
#
# [Parameters]
#     None
# [Return value]
#     Number of pages in the menu

sub num_pages()
{
    my ($self) = @_;
    return scalar(@{$self->{pages}});
}

###########################################################################

# generate_page:  Generate the M2TS stream for one page of the menu.
#
# [Parameters]
#             $page: Number of page to generate (0 = first page)
#        $m2ts_base: M2TS file number of first page
#     $menu_program: Menu program number
# [Return value]
#     M2TS stream data, as a string

sub generate_page($$$)
{
    my ($self, $page, $m2ts_base, $menu_program) = @_;

    # Technically, we could have a separate palette per menu page, but
    # be lazy for now.  Might be better to drop all this stuff into a
    # BDMenuPage class and just have BDMenu wrap that.
    if (!$self->{palette_map}) {
        $self->make_palette() or die "too many colors";
    }

    my $video_pid = 0x1011;
    my $menu_pid = 0x1400;

    my @video_packets = $self->generate_video($page);
    my @menu_packets =
        $self->generate_control($page, $m2ts_base, $menu_program);

    sub make_xts($$) {
        my ($type,$xts) = @_;  # $type == 1..3
        return pack("Cnn",
                    $type << 4 | int($xts / (1<<30)) << 1 | 1,
                    (($xts % (1<<30)) >> 15) << 1 | 1,
                    (($xts % (1<<30)) & 0x7FFF) << 1 | 1);
    }

    # I don't know whether all this complex PTS/DTS calculation is
    # necessary for simple menus like these, but leaving it in for now.

    my $initial_pcr = 27000000;  # FIXME: from M2TSGenerator.pm
    my $initial_pts = int($initial_pcr / 300) + int(0.25 * 90000);  # Should be enough leeway
    my $pts_inc = 450;  # Not sure how much this matters, but increment by 5msec per packet
    my $frame_inc = int(90000 * (1.001/24));
    my $pts = $initial_pts;
    foreach my $packet (@menu_packets) {
        my $type = unpack("C", substr($packet,0,1));
        if ($type == PRIVATE_STREAM_GRAPHICS) {
            $pts += $pts_inc;
        }
    }
    my $final_pts = $pts;
    if ($final_pts - $initial_pts < $frame_inc) {
        my $diff = $final_pts - $initial_pts;
        $final_pts = $initial_pts + $frame_inc;
        $initial_pts = $final_pts - $diff;
    }

    my $pts = $initial_pts;
    foreach my $packet (@menu_packets) {
        my $type = unpack("C", substr($packet,0,1));
        my $xts_flags;
        my $xts_data;
        if ($type == PRIVATE_STREAM_CONTROL) {
            $xts_flags = 0xC0;
            $xts_data = &make_xts(3, $final_pts) . &make_xts(1, $initial_pts);
        } elsif ($type == PRIVATE_STREAM_GRAPHICS) {
            my $dts = $pts;
            $pts += $pts_inc;
            $xts_flags = 0xC0;
            $xts_data = &make_xts(3, $pts) . &make_xts(1, $dts);
        } else {
            $xts_flags = 0x80;
            $xts_data = &make_xts(2, $pts);
        }
        $packet = pack("CCC", 0x84, $xts_flags, length($xts_data))
                . $xts_data . $packet;
        $packet = pack("Nn", 0x1BD, length($packet)) . $packet;
    }

    my $pts = $final_pts - $frame_inc;
    foreach my $packet (@video_packets) {
        my $dts = $pts;
        $pts += $frame_inc;
        my $xts_data = &make_xts(3, $pts) . &make_xts(1, $dts);
        $packet = pack("NnCCC", 0x1E0, 0, 0x80, 0xC0, length($xts_data))
                . $xts_data . $packet;
    }

    my $m2ts_gen = M2TSGenerator->create();
    $m2ts_gen->push_packet($video_pid, @video_packets);
    $m2ts_gen->push_packet($menu_pid, @menu_packets);
    return $m2ts_gen->generate();
}

###########################################################################
########################### Internal functions ############################
###########################################################################

# generate_video:  Generate the video data for the menu.
#
# [Parameters]
#     $page: Number of page to generate (0 = first page)
# [Return value]
#     Array of PES packets comprising the video stream

sub generate_video($)
{
    my ($self, $page) = @_;

    # FIXME: insecure temp files
    my $temp_ppm = "/tmp/bd-gen.ppm";
    my $temp_y4m = "/tmp/bd-gen.y4m";
    my $temp_m2v = "/tmp/bd-gen.m2v";

    $self->{pages}[$page]{background} ||= [0x000000];
    my ($bg_color, $bg_image, $bg_resize_type, $bg_resize_keep_aspect)
        = @{$self->{pages}[$page]{background}};
    my $bg_r = $bg_color>>16 & 0xFF;
    my $bg_g = $bg_color>> 8 & 0xFF;
    my $bg_b = $bg_color>> 0 & 0xFF;

    my $gimp_script = "
(let* ((outfile \"$temp_ppm\")
       (width 1920)
       (height 1080)
       (image (car (gimp-image-new width height RGB)))
       (layer (car (gimp-layer-new image width height RGBA-IMAGE \"layer\" 100
                                   NORMAL-MODE))))
  (gimp-image-add-layer image layer -1)
  (gimp-context-set-background '($bg_r $bg_g $bg_b))
  (gimp-drawable-fill layer BACKGROUND-FILL)
";
    if ($bg_image) {
        $gimp_script .= "
  (let* ((bgimage-layer (car (gimp-file-load-layer RUN-NONINTERACTIVE image \"$bg_image\")))
         (old-width (car (gimp-drawable-width bgimage-layer)))
         (old-height (car (gimp-drawable-height bgimage-layer)))
         (aspect-new-width (round (* old-width (/ height old-height))))
         (aspect-new-height (round (* old-height (/ width old-width))))
         (new-width old-width)
         (new-height old-height))
    (gimp-image-add-layer image bgimage-layer -1)
";
        if ($bg_resize_type) {
            if ($bg_resize_keep_aspect) {
                $gimp_script .= "
    (if (> aspect-new-width width)
        (begin
          (set! new-width width)
          (set! new-height aspect-new-height))
      (begin
        (set! new-width aspect-new-width)
        (set! new-height height)))
";
            } else {
                $gimp_script .= "
    (set! new-width width)
    (set! new-height height)
";
            }
            my %gimp_resize_types = (
                linear => 'INTERPOLATION-LINEAR',
                cubic  => 'INTERPOLATION-CUBIC',
                sinc   => 'INTERPOLATION-LANCZOS',
            );
            my $type = $gimp_resize_types{$bg_resize_type};
            $gimp_script .= "
    (unless (and (= old-width width) (= old-height height))
        (gimp-layer-scale-full bgimage-layer new-width new-height TRUE $type))
";
        }
        $gimp_script .= "
    (gimp-layer-set-offsets bgimage-layer (round (/ (- width new-width) 2))
                                          (round (/ (- height new-height) 2)))
    (set! layer (car (gimp-image-flatten image))))
";
    }
    foreach my $textinfo (@{$self->{pages}[$page]{text}}) {
        my ($x, $y, $align, $font, $size, $color, $text) = @$textinfo;
        $font =~ s/\"/\\\"/g;
        $text =~ s/\"/\\\"/g;
        my ($a, $r, $g, $b) = unpack("C*", pack("N", $color));
        $gimp_script .= "
  (let* ((font \"$font\")
         (size $size)
         (text \"$text\")
         (line-extents (gimp-text-get-extents-fontname text size PIXELS font))
         (line-width (car line-extents))
         (line-ascent (caddr line-extents))
         (x (- $x (/ (* line-width ".($align+1).") 2)))
         (y (- $y line-ascent)))
    (gimp-context-set-foreground '($r $g $b))
    (let* ((text-layer (car (gimp-text-fontname image -1 x y text 0 TRUE size PIXELS font))))
      (gimp-layer-set-opacity text-layer ".($a/255 * 100).")))
";
    }
    $gimp_script .= "
  (set! layer (car (gimp-image-flatten image)))
  (file-ppm-save RUN-NONINTERACTIVE image layer outfile outfile 1)
  (gimp-image-delete image)
  (gimp-quit 0)
)
";
    system "gimp", "-d", "-i", "-s", "-b", $gimp_script;
    die "gimp failed" if $? != 0;

    system "ppmtoy4m -S 420mpeg2 -n 18 -r -F 24000:1001 -A 1:1 '$temp_ppm'" .
           " > '$temp_y4m'";
    die "ppmtoy4m failed\n" if $? != 0;
    unlink $temp_ppm;

    system "mencoder", $temp_y4m, "-o", $temp_m2v, "-of", "rawvideo",
           "-ovc", "lavc", "-lavcopts", "vcodec=mpeg2video:cgop:sc_threshold=1000000000:vqscale=2:vmax_b_frames=0:keyint=18";
    die "mencoder failed\n" if $? != 0;
    unlink $temp_y4m;

    my $video_data;
    local *IN;
    open IN, "<$temp_m2v" or die "$temp_m2v: $!\n";
    binmode IN;
    {
        local $/ = undef;
        $video_data = <IN>;
    }
    close IN;
    unlink $temp_m2v;

    # Just treat the whole video stream as a single PES packet, because
    # we're lazy and because it seems to work anyway
    return ($video_data);
}

###########################################################################

# generate_control:  Generate the control data for the menu.
#
# [Parameters]
#             $page: Number of page to generate (0 = first page)
#        $m2ts_base: M2TS file number of first page
#     $menu_program: Menu program number
# [Return value]
#     Array of PES packets comprising the control stream

sub generate_control($$$)
{
    my ($self, $page, $m2ts_base, $menu_program) = @_;

    # Keep track of the images used in this page, so we can leave out any
    # that are only used in other pages.
    my %images_used = ();

    my @packets = ();

    my $num_entries = 0;
    foreach my $link (@{$self->{pages}[$page]{links}}) {
        $num_entries++ if defined($link);
    }
    my $packet = pack("NNNNNNNnNn",
                      0, 0, 0, 0x100, 0, 0, 0, 0, -1, $num_entries);
    for (my $id = 0; $id < scalar(@{$self->{pages}[$page]{links}}); $id++) {
        next if !defined($self->{pages}[$page]{links}[$id]);
        my ($x, $y, $normal, $hilite, $active,
            $up, $down, $left, $right, $auto, $type, $target, $chapter) =
                @{$self->{pages}[$page]{links}[$id]};
        $images_used{$normal} = 1 if $normal >= 0;
        $images_used{$hilite} = 1 if $hilite >= 0;
        $images_used{$active} = 1 if $active >= 0;
        $packet .= pack("nCnnCnnnnnnnnnnnnnn",
                        $id, 1, $id,
                        $type eq 'PAGE' ? -1 : $id,
                        $auto ? 0x80 : 0,
                        $x, $y, $up, $down, $left, $right,
                        # FIXME: Not sure why these have to be given twice,
                        # or what the third parameter is, but the 0xFF
                        # values at least don't seem to be placeholders
                        # for command lists.
                        $normal, $normal, 0xFF,
                        $hilite, $hilite, 0xFF,
                        $active, $active,
                       );
        if ($type eq 'PAGE') {
            $packet .= pack("nN*", 2,
                            0x50400001, REG_cur_menu, $m2ts_base + $target,
                            0x21810000, $menu_program, 0
                           );
        } else {
            $packet .= pack("nN*", 4,
                            0x50400001, REG_cur_menu, $m2ts_base + $page,
                            0x50400001, REG_cur_title, $target,
                            0x50400001, REG_chapter, $chapter,
                            0x21010000, REG_cur_title, 0
                           );
        }
    }
    $packet = "\x07\x80\x04\x38\x40\x00\x00\x80\xC0\x00"
            . pack("n", length($packet)) . $packet;
    $packet = pack("Cn", PRIVATE_STREAM_CONTROL, length($packet)) . $packet;
    push @packets, $packet;

    # This packet seems to specify the color palette in YVUA (Y/Cr/Cb/A)
    # format.  I guess color 0 is assumed to be transparent?
    $packet = "\x00\x00";
    for (my $i = 0; $i < @{$self->{palette}}; $i++) {
        $packet .= pack("CN", $i+1, $self->{palette}[$i]);
    }
    $packet = pack("Cn", PRIVATE_STREAM_PALETTE, length($packet)) . $packet;
    push @packets, $packet;

    foreach my $i (sort(keys(%images_used))) {
        $packet = $self->encode_image($self->{images}[$i]);
        # The first value here is the image number for menu commands.
        $packet = pack("nnCn", $i, 0xC0, 0, length($packet)) . $packet;
        $packet = pack("Cn", PRIVATE_STREAM_GRAPHICS, length($packet))
                . $packet;
        push @packets, $packet;
    }

    $packet = "";
    $packet = pack("Cn", PRIVATE_STREAM_TERMINATOR, length($packet)) . $packet;
    push @packets, $packet;

    return @packets;
}

###########################################################################

# make_palette:  Generate a color palette for all loaded images.
#
# [Parameters]
#     None
# [Return value]
#     True on success, false if there were too many colors (>255)

sub make_palette()
{
    my ($self) = @_;

    use integer;

    $self->{palette} = [];
    $self->{palette_map} = ();

    foreach my $image (@{$self->{images}}) {
        foreach my $row (@{$image->{data}}) {
            foreach my $pixel (@$row) {
                next if $pixel == 0;
                if (!defined($self->{palette_map}{$pixel})) {
                    if (scalar(@{$self->{palette}}) >= 255) {
                        return undef;
                    }
                    $self->{palette_map}{$pixel} = scalar(@{$self->{palette}}) + 1;
                    push @{$self->{palette}}, $pixel;
                }
            }
        }
    }

    return 1;
}

###########################################################################

# encode_image:  Encode the given image as 8-bit indexed RLE data.
#
# [Parameters]
#     $image: Image to encode
# [Return value]
#     Encoded data, as a string

sub encode_image($)
{
    my ($self, $image) = @_;

    use integer;

    my $data = pack("nn", $image->{width}, $image->{height});

    for (my $y = 0; $y < $image->{height}; $y++) {
        my $last_pixel = -1;
        my $last_count = 0;
        # Deliberately overruns the width by 1, so we don't have to
        # duplicate the RLE termination check
        for (my $x = 0; $x <= $image->{width}; $x++) {
            my $pixel = $image->{data}[$y][$x];
            if (defined($pixel) && $pixel != 0) {
                $pixel = $self->{palette_map}{$pixel};
            }
            if (defined($pixel) && $pixel == $last_pixel) {
                $last_count++;
            } else {
                if ($last_pixel != -1) {
                    if ($last_pixel == 0) {
                        $data .= "\0";
                        if ($last_count >= 0x40) {
                            $data .= pack("n", 0x4000 | $last_count);
                        } else {
                            $data .= pack("C", $last_count);
                        }
                    } elsif ($last_count > 2) {
                        $data .= "\0";
                        if ($last_count >= 0x40) {
                            $data .= pack("n", 0xC000 | $last_count);
                        } else {
                            $data .= pack("C", 0x80 | $last_count);
                        }
                        $data .= pack("C", $last_pixel);
                    } else {
                        $data .= pack("C", $last_pixel) x $last_count;
                    }
                }
                $last_pixel = $pixel;
                $last_count = 1;
            }
        }
        $data .= "\0\0";
    }

    return $data;
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

# pages
#     List of menu pages.  Each element is an hash reference containing:
#         'background => array containing data about the menu page background
#         'text'  => array of arrays containing data about text strings to draw
#         'links' => array of arrays containing data about links
# images
#     List of images used in the menu.  Each element is a hash reference
#     containing 'width', 'height', and 'data' fields.  The image data for
#     each image is stored in Y/Cr/Cb/A format.
# palette
#     Color palette for menu graphics (list of 0-255 entries), in Y/Cr/Cb/A
#     format.
# palette_map
#     Map (hash) from pixel values to palette indices.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
