# Module for creating a simple (text-only) Blu-ray Disc menu
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package BDSimpleMenu;

use strict;
use BDMenu;
use PNGReader;

###########################################################################

use constant PRIVATE_STREAM_PALETTE    => 0x14;  # Color palette data
use constant PRIVATE_STREAM_GRAPHICS   => 0x15;  # Menu graphic data
use constant PRIVATE_STREAM_CONTROL    => 0x18;  # Menu control data
use constant PRIVATE_STREAM_TERMINATOR => 0x80;  # I'll Be Bach

# Note that all coordinates are on a 1920x1080 frame
use constant TITLE_LIST_X          =>  280;
use constant TITLE_LIST_Y          =>  240;
use constant TITLE_LIST_LINESIZE   =>   60;
use constant TITLE_LIST_FONTSIZE   =>   50;
use constant TITLE_LIST_ARROW_XOFS =>  -50;
use constant TITLE_LIST_ARROW_YOFS =>  -34;
use constant TITLE_LIST_MAX_LINES  =>   12;  # Maximum per page
use constant TITLE_LIST_PAGENUM_X  => 1700;  # Right edge of text
use constant TITLE_LIST_PAGENUM_Y  =>  880;

###########################################################################

# Normal (menu item highlighted) arrow graphic data
my $arrow_normal_png =
    "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A\x00\x00\x00\x0D\x49\x48\x44\x52" .
    "\x00\x00\x00\x20\x00\x00\x00\x1F\x08\x06\x00\x00\x00\x86\x56\xCF" .
    "\x8C\x00\x00\x00\x01\x73\x52\x47\x42\x00\xAE\xCE\x1C\xE9\x00\x00" .
    "\x00\xA5\x49\x44\x41\x54\x48\xC7\xDD\x97\xB1\x0D\x02\x31\x10\x04" .
    "\xF7\xBE\x0F\x20\x25\x46\xA2\x7E\xE2\xA7\x0B\xBE\x03\x1A\x60\x08" .
    "\xC0\x92\xF5\x24\x48\x2F\x79\x10\xD7\x80\x47\x63\xFB\xEE\x36\xD9" .
    "\x58\xC0\x09\x38\xC6\x2C\xE0\x0A\xEC\x4C\x80\x33\xAF\xDA\x9B\x10" .
    "\xB3\x0A\xD1\x59\x50\x21\x66\xE0\xA1\x41\x74\x16\x54\x88\x66\xC1" .
    "\x81\x58\xBD\x05\x0D\xA2\x7F\x0B\xE3\x21\x56\x16\x34\x88\xCB\x37" .
    "\x10\x05\x30\xF2\x76\x92\x54\x92\x43\x55\x2D\x49\x32\x0D\xFE\x24" .
    "\xF5\x86\xB8\x35\x13\xA3\x0D\x7C\x9A\x90\x00\x1A\xC4\x5D\x37\x30" .
    "\xA9\xFA\xAB\x96\xC9\x3C\xFC\x27\xFA\xC0\x7F\x76\x42\x75\x16\xE8" .
    "\xD3\x50\xDD\x07\xF4\x8D\x48\xDD\x09\xF5\xAD\x58\xCD\x05\x7A\x32" .
    "\x52\xB3\xE1\xD6\x74\xFC\x04\xFF\x14\xAC\x60\x7D\xB4\x6D\xBE\x00" .
    "\x00\x00\x00\x49\x45\x4E\x44\xAE\x42\x60\x82";

# Active (menu item activated) arrow graphic data
my $arrow_active_png =
    "\x89\x50\x4E\x47\x0D\x0A\x1A\x0A\x00\x00\x00\x0D\x49\x48\x44\x52" .
    "\x00\x00\x00\x20\x00\x00\x00\x1F\x08\x06\x00\x00\x00\x86\x56\xCF" .
    "\x8C\x00\x00\x00\x01\x73\x52\x47\x42\x00\xAE\xCE\x1C\xE9\x00\x00" .
    "\x01\x11\x49\x44\x41\x54\x48\xC7\xC5\x97\x41\x0E\x82\x30\x10\x45" .
    "\x3F\x8D\xD7\x10\x27\x5D\xB9\x33\x31\xD1\xB5\xF7\x91\xE3\xE0\x79" .
    "\x74\x8D\x2B\x2F\x40\x46\x3C\x08\x6E\xDA\xA4\x81\x00\xAD\xCC\xD8" .
    "\x6E\x48\x48\xC8\x7F\xF3\x66\x5A\x00\x58\xB9\x98\xEC\x91\xC9\xEE" .
    "\x91\x73\x31\xD9\x86\xC9\x6E\x7F\x79\xD6\x08\x31\x5C\x01\x74\x4C" .
    "\xB6\xCC\x02\x40\xDC\x3E\x01\x34\x00\xDE\xA9\x10\x46\xB0\x13\x95" .
    "\xBB\x26\x41\x88\x01\x04\x16\xFA\x14\x08\x23\x3C\x8F\x15\x80\x22" .
    "\x05\x42\x14\x20\xB0\x80\x58\x08\xA3\xB0\x2B\xBD\x85\x28\x13\xE2" .
    "\x00\x83\x59\x58\x84\x30\x4A\x67\x93\xB7\x80\x25\x08\x15\x00\x67" .
    "\xE1\x11\xDC\x9A\x84\x28\x98\x6C\xFF\xC7\x53\xDB\xB7\x65\x47\xDC" .
    "\x76\x9A\x2D\x98\x5A\x23\x13\xFF\x36\x30\x32\x61\x90\x67\x79\x13" .
    "\xAF\x5C\x00\xDE\xC0\xC1\xE4\xD4\x4F\xDC\x76\x26\x67\x38\x82\xC3" .
    "\x42\xE3\x2B\xE9\x0E\xE0\x32\x17\xAE\x06\xC0\x64\x4F\xC1\x4B\x69" .
    "\x32\x5C\xF3\x1C\xA8\x5D\xE8\x6C\xB8\x8A\x81\x41\xF5\xB3\xE1\x00" .
    "\xB0\x51\xAE\x7E\x36\x5C\xBC\x05\xAE\xFA\x73\x6C\xB8\xC6\x0C\xD4" .
    "\x31\xDA\x55\x66\x60\xD0\xFB\xA8\x70\x69\x03\x75\x6A\xB8\x98\x81" .
    "\xA0\xFA\xA4\x70\xC9\x5D\x70\x03\x50\x12\xB7\x9F\x1C\x3F\xA6\xAB" .
    "\xFE\x8E\xBF\xA3\x56\x86\x02\xF0\xF4\x18\x32\x00\x00\x00\x00\x49" .
    "\x45\x4E\x44\xAE\x42\x60\x82";

###########################################################################
############################# Class interface #############################
###########################################################################

# create:  Create a new simple menu instance.
#
# [Parameters]
#     None
# [Return value]
#     New menu instance

sub create()
{
    my ($class) = @_;

    my $self = {};
    bless $self, $class;

    $self->{menu} = BDMenu->create();

    $self->{page} = 0;
    $self->{line} = 0;

    my $png_arrow_normal = PNGReader->parse(\$arrow_normal_png) or die;
    my $png_arrow_active = PNGReader->parse(\$arrow_active_png) or die;
    $self->{image_arrow_normal} =
        $self->{menu}->png_to_image($png_arrow_normal);
    $self->{image_arrow_active} =
        $self->{menu}->png_to_image($png_arrow_active);

    return $self;
}

###########################################################################

# add_title:  Add a title to a simple menu.
#
# [Parameters]
#     $title: Title index
#      $name: Title text to add
# [Return value]
#     None

sub add_title($$)
{
    my ($self, $title, $name) = @_;

    if ($self->{line} >= TITLE_LIST_MAX_LINES) {
        $self->{menu}->add_link_to_page($self->{page}, 0x101,
                                        0, 0, undef, undef, undef,
                                        TITLE_LIST_MAX_LINES, 1, 1, 1,
                                        1, $self->{page} + 1);
        $self->{page}++;
        $self->{line} = 0;
        $self->{menu}->add_link_to_page($self->{page}, 0x100,
                                        0, 0, undef, undef, undef,
                                        TITLE_LIST_MAX_LINES, 1, 1, 1,
                                        1, $self->{page} - 1);
    }
    my $x = TITLE_LIST_X;
    my $y = TITLE_LIST_Y + TITLE_LIST_LINESIZE * $self->{line};
    $self->{menu}->add_text($self->{page}, $x, $y, -1,
                            "Sans", TITLE_LIST_FONTSIZE, 0xFFFFFFFF, $name);
    $self->{menu}->add_link_to_title($self->{page},
                                     $self->{line},
                                     $x + TITLE_LIST_ARROW_XOFS,
                                     $y + TITLE_LIST_ARROW_YOFS,
                                     undef,
                                     $self->{image_arrow_normal},
                                     $self->{image_arrow_active},
                                     $self->{line}==0 ? TITLE_LIST_MAX_LINES-1 : $self->{line} - 1,
                                     $self->{line}==TITLE_LIST_MAX_LINES-1 ? 0 : $self->{line} + 1,
                                     0x100,
                                     0x101,
                                     0, $title);
    $self->{line}++;
}

###########################################################################

# num_pages:  Return the number of pages in the menu.
#
# [Parameters]
#     None
# [Return value]
#     Number of pages in the menu

sub num_pages()
{
    my ($self) = @_;
    return $self->{page} + 1;
}

###########################################################################

# generate_page:  Generate the M2TS stream for one page of the menu.
# Calling this function finalizes the menu; attempting to add any further
# titles will have undefined behavior.
#
# [Parameters]
#             $page: Number of page to generate (0 = first page)
#        $m2ts_base: M2TS file number of first page
#     $menu_program: Menu program number
# [Return value]
#     M2TS stream data, as a string

sub generate_page($$$)
{
    my ($self, $page, $m2ts_base, $menu_program) = @_;

    if (!$self->{finalized}) {
        # Update up/down wraparound links on last page
        $self->{menu}->set_link_movement($self->{page}, $self->{line}-1,
                                         undef, 0, undef, undef);
        $self->{menu}->set_link_movement($self->{page}, 0,
                                         $self->{line}-1, undef, undef, undef);
        # Add left/right page wraparound links if a multi-page menu
        if ($self->{page} > 0) {
            $self->{menu}->add_link_to_page($self->{page}, 0x101,
                                            0, 0, undef, undef, undef,
                                            TITLE_LIST_MAX_LINES, 1, 1, 1,
                                            1, 0);
            $self->{menu}->add_link_to_page(0, 0x100,
                                            0, 0, undef, undef, undef,
                                            TITLE_LIST_MAX_LINES, 1, 1, 1,
                                            1, $self->{page});
        }
        # Add page number text if a multi-page menu
        if ($self->{page} > 0) {
            for (my $page = 0; $page <= $self->{page}; $page++) {
                $self->{menu}->add_text($page, TITLE_LIST_PAGENUM_X,
                                        TITLE_LIST_PAGENUM_Y, +1,
                                        "Sans", int(TITLE_LIST_FONTSIZE*2/3),
                                        0xFFFFFFFF,
                                        ($page+1) . "/" . ($self->{page}+1));
            }
        }
        # Finalization done
        $self->{finalized} = 1;
    }

    return $self->{menu}->generate_page($page, $m2ts_base, $menu_program);
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

# page
#     Current page to which title names are being added.
# line
#     Line number in current page of next title name to add.
# finalized
#     True if the menu has been finalized.
# image_arrow_normal, image_arrow_active
#     Image indices for the internal arrow graphics.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
