# M2TS file writer
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package M2TS;

use strict;
use Fcntl qw(SEEK_SET SEEK_CUR);

# FIXME: these ought to be declared in a common file
use constant STREAM_TYPE_MPEG2 => 0x02;
use constant STREAM_TYPE_AAC   => 0x0F;
use constant STREAM_TYPE_H264  => 0x1B;
use constant STREAM_TYPE_LPCM  => 0x80;
use constant STREAM_TYPE_AC3   => 0x81;
use constant STREAM_TYPE_SUB   => 0x90;  # Subpicture stream(?)
use constant STREAM_TYPE_MENU  => 0x91;  # Menu control stream

###########################################################################
############################# Class interface #############################
###########################################################################

# create:  Create a new M2TS stream.
#
# [Parameters]
#     $path: Pathname of M2TS file to create
# [Return value]
#     New stream instance

sub create($)
{
    my ($class, $path) = @_;

    my $self = {};
    bless $self, $class;

    $self->{exclude_pids} = {};
    $self->{path} = $path;
    open $self->{file}, "+>$path" or die "$path: $!\n";
    binmode $self->{file};
    $self->{num_packets} = 0;
    $self->{pat_pid} = 0x0000;

    return $self;
}

###########################################################################

# exclude_pid:  Mark a PID for exclusion from the output stream.  Any
# packets with the given PID passed to write_packet() will be replaced with
# dummy packets, and any references to the PID in the PMT will be removed.
#
# [Parameters]
#     $pid: PID to be excluded from the output stream
# [Return value]
#     None

sub exclude_pid($)
{
    my ($self, $pid) = @_;

    $self->{exclude_pids}{$pid} = 1;
}

###########################################################################

# split:  Split the current M2TS stream at the most recent splittable
# point, returning the newly-created stream.  The split stream is closed
# and can no longer be written to.
#
# [Parameters]
#     $path: Pathname of new M2TS file to create
# [Return value]
#     New stream instance

sub split($)
{
    my ($self, $path) = @_;

    # Find a split point in the current stream
    if (scalar(@{$self->{keyframes}}) <= 1) {
        die "Failed to split $self->{path} (no split point available)\n";
    }
    my $keyframe_ref = pop @{$self->{keyframes}};
    my $split_packet = $keyframe_ref->{packet_num};

    # Read in all data from the split point onwards
    seek $self->{file}, $split_packet * 192, SEEK_SET
        or die "seek($,$split_packet): $!\n";
    my $move_packets = $self->{num_packets} - $split_packet;
    my $move_len = $move_packets * 192;
    my $move_data;
    read $self->{file}, $move_data, $move_len;
    die "read($self->{path}): Unexpected EOF\n"
        if length($move_data) != $move_len;

    # Truncate the current file at the split point
    truncate $self->{file}, $split_packet * 192
        or die "truncate($self->{path}): $!\n";

    # Looks like file sizes have to be a multiple of 2k, so pad the current
    # file out to a multiple of 32 packets = 6144 bytes
    $self->{num_packets} = $split_packet;
    if ($self->{num_packets} % 32 != 0) {
        seek $self->{file}, ($split_packet - 1) * 192, SEEK_SET
            or die "seek($self->{path},$split_packet): $!\n";
        my $timestamp;
        read $self->{file}, $timestamp, 192;  # Leave file pointer at EOF
        die "read($self->{path}): Unexpected EOF on timestamp\n"
            if length($timestamp) != 192;
        $timestamp = unpack("N", $timestamp);
        my $dummy_packet = "\x47\x1F\xFF\x10" . ("\xFF" x 184);
        while ($self->{num_packets} % 32 != 0) {
            print {$self->{file}} pack("N", ++$timestamp) . $dummy_packet;
            $self->{num_packets}++;
        }
    }

    # Close out the current stream
    close $self->{file};
    $self->{pts_end} = $keyframe_ref->{pts};
    $self->{pts_end_packet} = $split_packet;

    # Create the new stream and copy data from the old one
    my $last_stream = $self;
    $self = M2TS->create($path);
    $self->{cont_counter} = $keyframe_ref->{_cont_counter};
    $self->{streams}      = $last_stream->{streams};
    $self->{streaminfo}   = $last_stream->{streaminfo};
    $self->{pat_pid}      = $last_stream->{pat_pid};
    $self->{pmt_pid}      = $last_stream->{pmt_pid};
    $self->{nit_pid}      = $last_stream->{nit_pid};
    $self->{psi_data}     = $last_stream->{psi_data};
    $self->{pcr_pid}      = $last_stream->{pcr_pid};
    $self->{video_pid}    = $last_stream->{video_pid};
    $self->{audio_pid}    = $last_stream->{audio_pid};

    # Insert copies of the PAT/PMT/NIT at the beginning of the file
    # FIXME: assumes each data packet will fit in one TS packet
    my $timestamp = substr($move_data, 0, 4);
    foreach my $pid ($self->{pat_pid}, $self->{pmt_pid}, $self->{nit_pid}) {
        die sprintf("PID %04X data missing\n", $pid)
            if !defined($self->{psi_data}{$pid});
        my $packet = $timestamp . pack("CnC", 0x47, 0x4000 | $pid, 0x10)
                                . $self->{psi_data}{$pid};
        $self->write_packet($packet);
    }

    # Append the data split off from the previous file
    for (my $i = 0; $i < length($move_data); $i += 192) {
        $self->write_packet(substr($move_data, $i, 192));
    }

    # Return the new stream
    return $self;
}

###########################################################################

# close:  Close out the current M2TS stream.  Once closed, the stream can
# no longer be written to.
#
# [Parameters]
#     None
# [Return value]
#     None

sub close($)
{
    my ($self) = @_;

    # Looks like file sizes have to be a multiple of 2k, so pad the current
    # file out to a multiple of 32 packets = 6144 bytes
    if ($self->{num_packets} % 32 != 0) {
        seek $self->{file}, ($self->{num_packets} - 1) * 192, SEEK_SET
            or die "seek($self->{path},$self->{num_packets}): $!\n";
        my $timestamp;
        read $self->{file}, $timestamp, 192;  # Leave file pointer at EOF
        die "read($self->{path}): Unexpected EOF on timestamp\n"
            if length($timestamp) != 192;
        $timestamp = unpack("N", $timestamp);
        my $dummy_packet = "\x47\x1F\xFF\x10" . ("\xFF" x 184);
        while ($self->{num_packets} % 32 != 0) {
            print {$self->{file}} pack("N", ++$timestamp) . $dummy_packet;
            $self->{num_packets}++;
        }
    }

    # Close out the stream
    close $self->{file};
}

###########################################################################

# num_packets, pts_start, pts_end, pts_end_packet, streams, streaminfo,
# pmt_pid, pcr_pid, video_pid, audio_pid, keyframes:  Return stream data.
#
# [Parameters]
#     None
# [Return value]
#     Corresponding data field

sub num_packets()
{
    my ($self) = @_;
    return $self->{num_packets};
}

sub pts_start()
{
    my ($self) = @_;
    return $self->{pts_start};
}

sub pts_end()
{
    my ($self) = @_;
    return $self->{pts_end};
}

sub pts_end_packet()
{
    my ($self) = @_;
    return $self->{pts_end_packet};
}

sub streams()
{
    my ($self) = @_;
    return $self->{streams};
}

sub streaminfo() {
    my ($self) = @_;
    return $self->{streaminfo};
}

sub pmt_pid() {
    my ($self) = @_;
    return $self->{pmt_pid};
}

sub pcr_pid() {
    my ($self) = @_;
    return $self->{pcr_pid};
}

sub video_pid() {
    my ($self) = @_;
    return $self->{video_pid};
}

sub audio_pid() {
    my ($self) = @_;
    return $self->{audio_pid};
}

sub keyframes() {
    my ($self) = @_;
    return $self->{keyframes};
}

###########################################################################

# write_packet:  Write a single TS packet to the stream.
#
# [Parameters]
#     $packet: Packet to write:
# [Return value]
#     None

sub write_packet($)
{
    my ($self, $packet) = @_;
    my $packet_num = $self->{num_packets};

    # Parse the TS packet header
    my ($pid, $flags) = unpack("nC", substr($packet,5,3));
    my $is_start = (($pid & 0x4000) != 0);
    $pid &= 0x1FFF;
    my $cc = $flags & 0x0F;  # Continuity counter
    $flags &= 0xF0;

    # Drop the packet if it's just padding or it's on an excluded PID
    if (!($flags & 0x30) || defined($self->{exclude_pids}{$pid})) {
        return;
    }

    # Write the packet and skip out if there's only an adaptation field
    # (i.e., no data)
    if (!($flags & 0x10)) {
        print {$self->{file}} $packet;
        $self->{num_packets}++;
        return;
    }

    # Save the original packet data
    my $saved_packet = $packet;

    # Skip the M2TS timestamp, TS header, and any adaptation field
    my $adapt_len = ($flags & 0x20 ? 1 + unpack("C",substr($packet,8,1)) : 0);
    $packet = substr($packet, 8 + $adapt_len);

    # Check whether it's a PSI or data packet
    my $is_psi = (defined($self->{pat_pid}) && $pid == $self->{pat_pid})
              || (defined($self->{pmt_pid}) && $pid == $self->{pmt_pid})
              || (defined($self->{nit_pid}) && $pid == $self->{nit_pid});

    # Watch out for stream discontinuities in data streams.  For PSI
    # packets, just rewrite the continuity counter ourselves (since we
    # insert extra copies when splitting files).
    if ($pid != 0x1FFF  # Padding stream
     && defined($self->{cont_counter}{$pid})
     && $self->{cont_counter}{$pid} != $cc
    ) {
        if ($is_psi) {
            $cc = $self->{cont_counter}{$pid};
            substr($saved_packet, 7, 1) = pack("C", $flags | $cc);
        } else {
            die sprintf("TS discontinuity detected on PID 0x%04X at timestamp 0x%08X\n",
                        $pid, unpack("N", substr($saved_packet,0,4)));
        }
    }
    $self->{cont_counter}{$pid} = ($cc + 1) & 0xF;

    # Save the PAT/PMT/NIT data if we haven't already
    # FIXME: otherwise make sure it hasn't changed
    if ($is_psi && !defined($self->{psi_data}{$pid})) {
        $self->{psi_data}{$pid} = substr($saved_packet, 8);
    }

    # PSI packets (PAT, PMT, etc.) can be up to 1024 bytes long, split
    # among multiple TS packets.  The first TS packet has a leading byte
    # which apparently indicates how much of the TS packet belongs to the
    # previous control packet -- the same annoying overlap method used by
    # Ogg files.  That's a pain to deal with in the current code, so we
    # just bail if the packet-start pointer is nonzero.
    if ($is_psi && $is_start) {
        my $pointer = unpack("C", substr($packet,0,1));
        $packet = substr($packet, 1);
        if ($pointer > 0) {
            # TS packet also contains part of previous data packet
            die "Split PSI packets not supported at packet $packet_num\n";
        }
        $packet = substr($packet, $pointer);
        my $packet_len = 3 + (unpack("n", substr($packet,1,2)) & 0x0FFF);
        if ($packet_len < 0x400) {
            $packet = substr($packet, 0, $packet_len);
        }
    }

    # PSI packets seem to occur occasionally throughout a TS file, so only
    # process the first set and assume there are no changes later on.
    if ($pid == $self->{pat_pid} && !defined($self->{pmt_pid})) {
        $self->handle_pat($packet_num, $packet);

    } elsif (defined($self->{pmt_pid}) && $pid == $self->{pmt_pid}) {
        if (!defined($self->{video_pid})) {
            $self->handle_pmt($packet_num, $packet);
        }
        # The PMT may have been altered by PID exclusion, so rewrite it
        $packet = $self->{psi_data}{$pid};
        $saved_packet = substr($saved_packet,0,8) . $packet;

    # Remainder is data packet handling
    } elsif (defined($self->{streaminfo}{$pid})) {
        my $streaminfo = $self->{streaminfo}{$pid};
        if ($streaminfo->{type} == STREAM_TYPE_H264) {
            if (!defined($streaminfo->{format}) && $is_start) {
                $self->find_h264_params($streaminfo, $pid,
                                        $packet_num, $packet);
            }
        } elsif ($streaminfo->{type} == STREAM_TYPE_MPEG2) {
            if (!defined($streaminfo->{format}) && $is_start) {
                $self->find_mpeg2_params($streaminfo, $pid,
                                         $packet_num, $packet);
            }
        } elsif ($streaminfo->{type} == STREAM_TYPE_LPCM) {
            die "Don't know how to find LPCM parameters\n";
        } elsif ($streaminfo->{type} == STREAM_TYPE_AC3) {
            if (!defined($streaminfo->{rate}) && $is_start) {
                $self->find_ac3_params($streaminfo, $pid,
                                       $packet_num, $packet);
            }
        } elsif ($streaminfo->{type} == STREAM_TYPE_AAC) {
            if (!defined($streaminfo->{rate}) && $is_start) {
                $self->find_aac_params($streaminfo, $pid,
                                       $packet_num, $packet);
            }
        }

        if (defined($self->{video_pid}) && $pid == $self->{video_pid}) {
            if ($is_start || $self->{saved_vdata} ne "") {
                $self->{saved_vdata_packet_num} = $packet_num if $is_start;
                $self->{saved_vdata} .= $packet;
                if (length($self->{saved_vdata}) > 1000000) {
                    die "saved_vdata too large! (bug?)\n";
                }
                if ($self->{in_keyframe}) {
                    my $keyframe_ref =
                        $self->{keyframes}[scalar(@{$self->{keyframes}})-1];
                    $keyframe_ref->{size} =
                        $self->{keyframe_end}+1 - $keyframe_ref->{packet_num};
                    $self->{in_keyframe} = 0;
                }
                my @ptsdata = unpack("CCCCC", substr($self->{saved_vdata},9,5));
                my $pts_2 = ($ptsdata[0]>>1 & 0x07) << 29
                          | ($ptsdata[1]>>0 & 0xFF) << 21
                          | ($ptsdata[2]>>1 & 0x7F) << 14
                          | ($ptsdata[3]>>0 & 0xFF) <<  6
                          | ($ptsdata[4]>>1 & 0x7E) >>  1;
                my $pts_lowbit = ($ptsdata[4]>>1 & 0x01);
                my $pts = ($pts_2 * 2) + $pts_lowbit;
                if (!defined($self->{pts_start})) {
                    $self->{pts_start} = $pts;
                }
                $self->{pts_end} = $pts;
                my $is_keyframe = 0;
                if ($streaminfo->{type} == STREAM_TYPE_H264) {
                    for (my $i = 0; $i <= length($self->{saved_vdata})-4; $i++) {
                        if (substr($self->{saved_vdata},$i,3) eq "\0\0\1") {
                            my $type = unpack("C", substr($self->{saved_vdata},$i+3,1)) & 0x1F;
                            if ($type == 7 || $type == 5) {
                                $is_keyframe = 1;
                                $self->{saved_vdata} = "";
                                last;
                            } elsif ($type >= 1 && $type <= 4) {
                                # Not a keyframe
                                $self->{saved_vdata} = "";
                                last;
                            }
                        }
                    }
                } elsif ($streaminfo->{type} == STREAM_TYPE_MPEG2) {
                    for (my $i = 0; $i <= length($self->{saved_vdata})-4; $i++) {
                        if (substr($self->{saved_vdata},$i,4) eq "\0\0\1\xB8") {  # GOP
                            $is_keyframe = 1;
                            $self->{saved_vdata} = "";
                            last;
                        } elsif (substr($self->{saved_vdata},$i,4) eq "\0\0\1\0"  # Picture
                                 && $i < length($self->{saved_vdata})-6) {
                            my $type = unpack("C", substr($self->{saved_vdata},$i+5,1));
                            $type = ($type>>3) & 7;
                            if ($type == 1) {  # I-frame
                                $is_keyframe = 1;
                            }
                            $self->{saved_vdata} = "";
                            last;
                        }
                    }
                } else {
                    die "impossible: type=$streaminfo->{type}";
                }
                if ($is_keyframe) {
                    my %temp_cc = (%{$self->{cont_counter}});
                    $temp_cc{$pid} = ($temp_cc{$pid} - 1) & 0xF;
                    push @{$self->{keyframes}},
                    {pts => $pts,
                     packet_num => $self->{saved_vdata_packet_num},
                     _cont_counter => {%temp_cc}};
                    $self->{in_keyframe} = 1;
                }
            }  # $is_start
            $self->{pts_end_packet} = $packet_num;
            if ($self->{in_keyframe}) {
                $self->{keyframe_end} = $packet_num;
            }
        }

    }

    # Write the (possibly modified) packet
    print {$self->{file}} $saved_packet;
    $self->{num_packets}++;

}  # write_packet()

###########################################################################
########################### Internal functions ############################
###########################################################################

# handle_pat:  Process a TS PAT packet.  Helper for write_packet().

sub handle_pat($$)
{
    my ($self, $packet_num, $packet) = @_;

    $packet = substr($packet, 0, -4);  # Ignore CRC

    my ($tid, undef, $id, $version, $sec_num, $last_sec_num) =
        unpack("CnnCCC", $packet);
    $packet = substr($packet, 8);
    die "Invalid TID in PAT at packet $packet_num\n"
        if $tid != 0x00;  # TID_PAT

    while (length($packet) >= 4) {
        my ($sid, $sub_pid) = unpack("nn", $packet);
        $packet = substr($packet, 4);
        $sub_pid &= 0x1FFF;
        if ($sid == 0) {
            $self->{nit_pid} = $sub_pid;
        } else {
            die "Too many pmt_pids at packet $packet_num\n"
                if defined($self->{pmt_pid});
            $self->{pmt_pid} = $sub_pid;
        }
    }
}

###########################################################################

# handle_pmt:  Process a TS PMT packet.  Helper for write_packet().

sub handle_pmt($$)
{
    my ($self, $packet_num, $packet) = @_;
    my $new_packet = "";

    $packet = substr($packet, 0, -4);  # Ignore CRC

    my ($tid, undef, $id, $version, $sec_num, $last_sec_num) =
        unpack("CnnCCC", $packet);
    $new_packet .= substr($packet, 0, 8);
    $packet = substr($packet, 8);
    die "Invalid TID in PMT at packet $packet_num\n"
        if $tid != 0x02;  # TID_PMT

    my $pcr_pid = unpack("n", $packet);
    $new_packet .= substr($packet, 0, 2);
    $packet = substr($packet, 2);
    $pcr_pid &= 0x1FFF;
    $self->{pcr_pid} = $pcr_pid;
    my $program_info_length = unpack("n", $packet);
    $new_packet .= substr($packet, 0, 2);
    $packet = substr($packet, 2);
    $program_info_length &= 0x0FFF;
    my $program_info = substr($packet, 0, $program_info_length);
    $new_packet .= $program_info;
    $packet = substr($packet, $program_info_length);

    # Not sure what this is needed for, but let's watch for it
    my $has_hdmv_descr = 0;
    while (length($program_info) >= 2) {
        my ($tag, $len) = unpack("CC", $program_info);
        $program_info = substr($program_info, 2);
        if ($len > length($program_info)) {
            die "Broken PMT (program info) at packet $packet_num\n";
        }
        if ($tag == 0x05  # REGISTRATION_DESCRIPTOR
         && $len >= 4
         && substr($program_info, 0, 4) eq "HDMV"
        ) {
            $has_hdmv_descr = 1;
        }
        $program_info = substr($program_info, $len);
    }

    while (length($packet) >= 5) {
        my ($stream_type, $stream_pid, $desc_list_len) =
            unpack("Cnn", $packet);
        my $temp = substr($packet, 0, 5);
        $packet = substr($packet, 5);
        $stream_pid &= 0x1FFF;
        $desc_list_len &= 0x0FFF;
        if ($desc_list_len > length($packet)) {
            die "Broken PMT (stream list) at packet $packet_num\n";
        }
        my $lang = undef;
        my $desc_list = substr($packet, 0, $desc_list_len);
        $packet = substr($packet, $desc_list_len);
        next if $self->{exclude_pids}{$stream_pid};
        $new_packet .= $temp . $desc_list;
        while (length($desc_list) >= 1) {
            my ($tag, $len) = unpack("CC", $desc_list);
            $desc_list = substr($desc_list, 2);
            if ($len > length($desc_list)) {
                die "Broken PMT (stream info) at packet $packet_num\n";
            }
            if ($tag == 0x0A && $len >= 3 && $desc_list =~ /^[a-z]{3}/) {
                $lang = substr($desc_list, 0, 3);
            }
            $desc_list = substr($desc_list, $len);
        }
        push @{$self->{streams}}, $stream_pid;
        $self->{streaminfo}{$stream_pid} =
            {type => $stream_type, lang => $lang};
        if ($stream_type == STREAM_TYPE_MPEG2
         || $stream_type == STREAM_TYPE_H264
        ) {
            $self->{video_pid} = $stream_pid;
        } elsif ($stream_type == STREAM_TYPE_LPCM
              || $stream_type == STREAM_TYPE_AC3
              || $stream_type == STREAM_TYPE_AAC
        ) {
            $self->{audio_pid} = $stream_pid;
        }
    }

    die "No H.264 or MPEG-2 stream found at packet $packet_num\n"
        if !defined($self->{video_pid});

    # Update the saved packet data with the (possibly modified) stream list
    $new_packet .= pack("N", M2TSGenerator->crc32($new_packet));
    $self->{psi_data}{$self->{pmt_pid}} =
        pack("C", 0) . $new_packet . ("\xFF" x (183-length($new_packet)));
}

###########################################################################

# find_h264_params:  Parse an H.264 SPS packet to find video parameters.
# Helper for write_packet().

my $bitpos;  # for use by get_bits(), etc.

sub find_h264_params($$$$)
{
    my ($self, $streaminfo, $pid, $packet_num, $packet) = @_;

    my $data = undef;
    for (my $i = 0; $i <= length($packet)-4; $i++) {
        if ((unpack("N", substr($packet,$i,4)) & ~0x60) == 0x107) {
            $data = substr($packet, $i+4);
            last;
        }
    }
    die "Can't find video parameters at packet $packet_num\n"
        if !defined($data);

    # Drop any emulation prevention bytes (note that NALs are prohibited
    # from containing any 00 00 03 xx sequence where the 03 is not an
    # emulation prevention byte)
    $data =~ s/\x00\x00\x03/\x00\x00/g;

    sub get_bits($ ) {
        my ($nbits) = @_;
        return 0 if !$nbits;
        my $value = unpack("N", substr($data, int($bitpos/8), 4));
        $value >>= ((32 - $bitpos%8) - $nbits);
        $value &= (1 << $nbits) - 1;
        $bitpos += $nbits;
        return $value;
    }
    sub get_golomb() {
        my $nbits = 0;
        $nbits++ while (!get_bits(1));
        return ((1<<$nbits) - 1) + get_bits($nbits);
    }
    sub get_sgolomb() {
        my $code = get_golomb();
        return $code & 1 ? $code>>1 : -($code>>1);
    }

    $bitpos = 0;
    my $profile_idc = get_bits(8);
    get_bits(8);
    my $level_idc = get_bits(8);
    my $sps_id = get_golomb();
    if ($profile_idc >= 100) {  # High profile parameters
        my $chroma_format_idc = get_golomb();
        if ($chroma_format_idc == 3) {
            my $residual_color_transform_flag = get_bits(1);
        }
        my $bit_depth_luma = 8 + get_golomb();
        my $bit_depth_chroma = 8 + get_golomb();
        my $transform_bypass = get_bits(1);
        my $scaling_matrix_present = get_bits(1);
        if ($scaling_matrix_present) {
            for (my $i = 0; $i < 8; $i++) {
                if (get_bits(1)) {
                    die "Scaling matrices not handled at packet $packet_num\n";
                }
            }
        }
    }
    my $log2_max_frame_num = 4 + get_golomb();
    my $poc_type = get_golomb();
    if ($poc_type == 0) {
        my $log2_max_poc_lsb = 4 + get_golomb();
    } elsif ($poc_type == 1) {
        my $delta_pic_order_always_zero_flag = get_bits(1);
        my $offset_for_non_ref_pic = get_sgolomb();
        my $offset_for_top_to_bottom_field = get_sgolomb();
        my $poc_cycle_length = get_golomb();
    } elsif ($poc_type != 2) {
        die "Invalid poc_type $poc_type at packet $packet_num\n";
    }
    my $ref_frame_count = get_golomb();
    my $gaps_in_frame_num_allowed_flag = get_bits(1);
    my $mb_width = 1 + get_golomb();
    my $mb_height = 1 + get_golomb();
    my $frame_mbs_only_flag = get_bits(1);
    my $mb_aff = $frame_mbs_only_flag ? 0 : get_bits(1);
    my $direct_8x8_inference_flag = get_bits(1);
    my $crop = get_bits(1);
    if ($crop) {
        my $crop_left = get_golomb();
        my $crop_right = get_golomb();
        my $crop_top = get_golomb();
        my $crop_bottom = get_golomb();
    }
    my $vui_parameters_present_flag = get_bits(1);
    my ($sar, $frame_rate);
    if ($vui_parameters_present_flag) {
        my $aspect_ratio_info_present_flag = get_bits(1);
        if ($aspect_ratio_info_present_flag) {
            my $aspect_ratio_idc = get_bits(8);
            if ($aspect_ratio_idc == 255) {
                my $sar_num = get_bits(16);
                my $sar_den = get_bits(16);
                $sar = $sar_num / $sar_den;
            } else {
                my @sar_table = (undef, 1/1, 12/11, 10/11,
                                 16/11, 40/33, 24/11, 20/11,
                                 32/11, 80/33, 18/11, 15/11,
                                 64/33, 160/99);
                $sar = $sar_table[$aspect_ratio_idc];
                die "Can't handle aspect_ratio_idc $aspect_ratio_idc at" .
                    " packet $packet_num\n"
                    if !defined($sar);
            }
        }
        my $overscan_info_present_flag = get_bits(1);
        my $overscan_appropriate_flag = $overscan_info_present_flag ? get_bits(1) : 0;
        my $video_signal_type_present_flag = get_bits(1);
        if ($video_signal_type_present_flag) {
            my $video_format = get_bits(3);
            my $video_full_range_flag = get_bits(1);
            my $colour_description_present_flag = get_bits(1);
            if ($colour_description_present_flag) {
                my $colour_primaries = get_bits(8);
                my $transfer_characteristics = get_bits(8);
                my $matrix_coefficients = get_bits(8);
            }
        }
        my $chroma_location_info_present_flag = get_bits(1);
        if ($chroma_location_info_present_flag) {
            my $chroma_sample_location_type_top_field = get_bits(1);
            my $chroma_sample_location_type_bottom_field = get_bits(1);
        }
        my $timing_info_present_flag = get_bits(1);
        if ($timing_info_present_flag) {
            my $num_units_in_tick = get_bits(16);
            $num_units_in_tick <<= 16;
            $num_units_in_tick |= get_bits(16);
            my $time_scale = get_bits(16);
            $time_scale <<= 16;
            $time_scale |= get_bits(16);
            $frame_rate = $time_scale / $num_units_in_tick;
            my $fixed_frame_rate_flag = get_bits(1);
        }
        # No need for the rest
    }

    my $aspect;
    if ($mb_width == 720/16 && $mb_height == 240/16) {
        $streaminfo->{format} = '480i';
        $aspect = ($sar ? (720/480) * $sar : 4/3);
        $frame_rate ||= 60/1.001;
        $frame_rate /= 2;
    } elsif ($mb_width == 720/16 && $mb_height == 480/16) {
        $streaminfo->{format} = '480p';
        $aspect = ($sar ? (720/480) * $sar : 4/3);
        $frame_rate ||= 30/1.001;
    } elsif ($mb_width == 720/16 && $mb_height == 288/16) {
        $streaminfo->{format} = '576i';
        $aspect = ($sar ? (720/576) * $sar : 4/3);
        $frame_rate ||= 50;
        $frame_rate /= 2;
    } elsif ($mb_width == 720/16 && $mb_height == 576/16) {
        $streaminfo->{format} = '576p';
        $aspect = ($sar ? (720/576) * $sar : 4/3);
        $frame_rate ||= 25;
    } elsif ($mb_width == 1280/16 && $mb_height == 720/16) {
        $streaminfo->{format} = '720p';
        $aspect = ($sar ? (1280/720) * $sar : 16/9);
        $frame_rate ||= 60/1.001;
    } elsif ($mb_width == 1440/16 && $mb_height == 544/16) {
        $streaminfo->{format} = '1080i';
        $aspect = ($sar ? (1440/1080) * $sar : 16/9);
        $frame_rate ||= 60/1.001;
        $frame_rate /= 2;
    } elsif ($mb_width == 1440/16 && $mb_height == 1088/16) {
        $streaminfo->{format} = '1080p';
        $aspect = ($sar ? (1440/1080) * $sar : 16/9);
        $frame_rate ||= 60/1.001;
    } elsif ($mb_width == 1920/16 && $mb_height == 544/16) {
        $streaminfo->{format} = '1080i';
        $aspect = ($sar ? (1920/1080) * $sar : 16/9);
        $frame_rate ||= 60/1.001;
        $frame_rate /= 2;
    } elsif ($mb_width == 1920/16 && $mb_height == 1088/16) {
        $streaminfo->{format} = '1080p';
        $aspect = ($sar ? (1920/1080) * $sar : 16/9);
        $frame_rate ||= 60/1.001;
    } else {
        die sprintf("Invalid frame format %dx%d%s at packet %d\n",
                    $mb_width*16, $mb_height*16,
                    $mb_aff ? 'i' : 'p', $packet_num);
    }
    $streaminfo->{widescreen} = ($aspect > 14/9);
    $streaminfo->{frame_rate} = int($frame_rate);
}

###########################################################################

# find_mpeg2_params:  Parse an MPEG-2 sequence header to find video
# parameters.  Helper for write_packet().

sub find_mpeg2_params($$$$)
{
    my ($self, $streaminfo, $pid, $packet_num, $packet) = @_;

    my $data = undef;
    for (my $i = 0; $i <= length($packet)-4; $i++) {
        if (unpack("N", substr($packet,$i,4)) == 0x1B3) {
            $data = substr($packet, $i+4, 8);
            die "MPEG-2 sequence header is split across packets at packet" .
                " $packet_num, cannot process\n"
                if length($data) < 4;
            last;
        }
    }
    die "Can't find video parameters at packet $packet_num\n"
        if !defined($data);

    my $width           = unpack("n", substr($data,0,2)) >> 4;
    my $height          = unpack("n", substr($data,1,2)) & 0xFFF;
    my $aspect_code     = unpack("C", substr($data,3,1)) >> 4;
    my $frame_rate_code = unpack("C", substr($data,3,1)) & 0xF;

    my @frc_table = (undef, 24/1.001, 24, 25, 30/1.001, 30, 50, 60/1.001, 60);
    my $frame_rate = $frc_table[$frame_rate_code];
    die "Bad frame rate code $frame_rate_code at packet $packet_num\n"
        if !defined($frame_rate);

    my @aspect_table = (undef, 1/1, 4/3, 16/9, 2.21/1);
    my $aspect = $aspect_table[$aspect_code];
    die "Bad aspect code $aspect_code at packet $packet_num\n"
        if !defined($aspect);

    my $progressive_char = undef;
    for (my $i = 0; $i <= length($packet)-4; $i++) {
        if (unpack("N", substr($packet,$i,4)) == 0x1B5
         && (unpack("C", substr($packet,$i+4,1)) >> 4) == 0x1
         && length($packet) >= $i+6
        ) {
            my $progressive = unpack("C", substr($packet,$i+5,1)) & (1<<3);
            $progressive_char = $progressive ? 'p' : 'i';
            last;
        }
    }
    if (!defined($progressive_char)) {
        $progressive_char = ($height <= 576) ? 'i' : 'p';
        print STDERR "WARNING: Can't find MPEG-2 sequence extension header, assuming %s video\n",
            $progressive_char eq 'p' ? 'progressive' : 'interlaced';
    }

    if ($width == 720 && $height == 480) {
        $streaminfo->{format} = '480' . $progressive_char;
    } elsif ($width == 720 && $height == 576) {
        $streaminfo->{format} = '576' . $progressive_char;
    } elsif ($width == 1280 && $height == 720 && $progressive_char eq 'p') {
        $streaminfo->{format} = '720p';
    } elsif (($width == 1440 || $width == 1920) && $height == 1080) {
        $streaminfo->{format} = '1080' . $progressive_char;
    } else {
        die sprintf("Invalid frame format %dx%d%s at packet $packet_num\n",
                    $width, $height, $progressive_char);
    }
    $streaminfo->{widescreen} = ($aspect > 14/9);
    $streaminfo->{frame_rate} = int($frame_rate);
}

###########################################################################

# find_ac3_params:  Parse an AC3 packet to find audio parameters.  Helper
# for write_packet().

sub find_ac3_params($$$$)
{
    my ($self, $streaminfo, $pid, $packet_num, $packet) = @_;

    if (substr($packet, 0, 3) eq "\0\0\1") {  # Skip over MPEG encapsulation
        $packet = substr($packet, 9 + unpack("C",substr($packet,8,1)));
    }
    if (substr($packet, 0, 2) ne "\x0B\x77") {
        die "Invalid/corrupt AC3 bitstream at packet $packet_num\n"
    }
    my $temp = unpack("N", substr($packet,4,4));
    if (($temp>>19 & 0x1F) > 10) {
        die "E-AC3 not supported at packet $packet_num\n";
    }
    $streaminfo->{rate} = (48000,44100,32000)[$temp>>30 & 3];
    if ($streaminfo->{rate} != 48000) {
        die "Sampling rate $streaminfo->{rate} invalid at packet $packet_num\n";
    }
    my $lfe_shift = (12,12,10,10,10,8,10,8)[$temp>>13 & 7];
    $streaminfo->{channels} = (2,1,2,3,3,4,4,5)[$temp>>13 & 7]
                              + ($temp>>$lfe_shift & 1);
}

###########################################################################

# find_aac_params:  Parse an AAC packet to find audio parameters.  Helper
# for write_packet().

sub find_aac_params($$$$)
{
    my ($self, $streaminfo, $pid, $packet_num, $packet) = @_;

    if (substr($packet, 0, 3) eq "\0\0\1") {  # Skip over MPEG encapsulation
        $packet = substr($packet, 9 + unpack("C",substr($packet,8,1)));
    }
    if ((unpack("n", substr($packet,0,2)) & ~0x1) != 0xFFF8) {
        die "Invalid/corrupt AAC bitstream at packet $packet_num\n"
    }
    my $temp = unpack("n", substr($packet,2,2));
    $streaminfo->{rate} = (96000,88200,64000,
                           48000,44100,32000,
                           24000,22050,16000,
                           12000,11025, 8000, 7350)[$temp>>10 & 15];
    if ($streaminfo->{rate} != 48000) {
        die "Sampling rate $streaminfo->{rate} invalid at packet $packet_num\n";
    }
    $streaminfo->{channels} = (0,1,2,3,4,5,6,8)[$temp>>6 & 7];
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

######## General control fields ########

# exclude_pids
#     Hash whose keys indicate PIDs to be excluded from the M2TS stream.
# path
#     The pathname of the file being written.
# file
#     The filehandle used for writing the M2TS stream.
# num_packets
#     The number of packets written to the M2TS stream.
# cont_counter
#     A hash containing the next continuity counter value for each PID.
# pts_start
#     The first PTS in the video stream.
# pts_end
#     The last PTS in the video stream.
# pts_end_packet
#     The packet index of the 'pts_end' packet.
# streams
#     An array of data stream PIDs, in the order found in the PMT.
# streaminfo
#     A hash of information about each data stream in the file.  Each entry
#     is itself a hash, containing the following keys:
#         type       => The TS type of the stream (0x1B = H.264, etc.)
#         lang       => The 3-character language code of the stream, or
#                          undef if unknown.
#         format     => (video only) The video format: '480i', '720p', etc.
#         frame_rate => (video only) The frame rate (frames/sec), truncated
#                          to an integer.
#         widescreen => (video only) True if a widescreen video, else false.
#         rate       => (audio only) The sampling rate (samples/sec).
#         channels   => (audio only) The number of channels in the stream.
# keyframes
#     An array of keyframes.  Each entry is a hash containing two keys:
#         pts        => The PTS of the keyframe.
#         packet_num => The packet index (starting from 0) within the M2TS
#                          stream.
#         size       => The size of the keyframe's data in packets,
#                          including any intervening packets from other PIDs.
# saved_vdata
#     Holds saved TS payload data from the beginning of a video frame's PES
#     packet until the frame is known either to be or not to be a keyframe.
# saved_vdata_packet_num
#     The packet index of the first TS packet contained in saved_vdata.
# in_keyframe
#     True from the beginning of a video keyframe's PES packet to the
#     beginning of the next video PES packet.
# keyframe_end
#     The packet index of the last TS packet associated with the current
#     video keyframe's PES packet.  Only used when in_keyframe is true.

######## PSI (Program Specific Information) and related data ########

# pat_pid
#     The PID of the Program Association Table (PAT).  Defined to be 0x0000
#     by the TS specification (ISO/IEC 13818-1), but included for
#     parallelism with the other PSI fields.
# pmt_pid
#     The PID of the Program Association Table (PAT).  Declared in the PAT.
# nit_pid
#     The PID of the Network Information Table (NIT).  Declared in the PAT.
# psi_data
#     A hash containing the PAT/PMT/NIT data, excluding the 32-bit M2TS
#     timestamp and 32-bit TS header, indexed by PID.
# pcr_pid
#     The PID of the Program Clock Reference (PCR).  Declared in the PMT.
# video_pid
#     The PID of the video stream.  Declared in the PMT.
# audio_pid
#     The PID of the audio stream.  Declared in the PMT.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
