# Module for reading image data from PNG files.  This module is only
# intended for use in the proof-of-concept Blu-ray disc structure generator
# "bd-gen", and as such is not particularly robust.
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

package PNGReader;

use strict;
use integer;
use Compress::Zlib;

###########################################################################
############################# Class interface #############################
###########################################################################

# read:  Read a PNG file and return a PNGReader object containing the
# file's image data.
#
# [Parameters]
#     path: Pathname of file to read
# [Return value]
#     PNGReader object containing image data, or undef on error

sub read($)
{
    my ($class, $path) = @_;
    local *F;

    if (!open(F, "<$path")) {
        print STDERR "$path: $!\n";
        return undef;
    }
    binmode F;
    undef $/;
    my $data = <F>;
    close F;

    return $class->parse(\$data, $path);
}

###########################################################################

# parse:  Parse a PNG file and return a PNGReader object containing the
# image data.  Used when the file has already been read into memory.
#
# [Parameters]
#     data_ref: Reference to PNG file data
#         path: Optional pathname used in error messages
# [Return value]
#     PNGReader object containing image data, or undef on error

sub parse($;$)
{
    my ($class, $data_ref, $path) = @_;
    $path = "$path: " if $path;

    my $self = {};
    bless $self, $class;

    if (substr($$data_ref, 0, 8) ne "\x89PNG\x0D\x0A\x1A\x0A") {
        print STDERR "${path}File is not a PNG file\n";
        return undef;
    }

    my $pos = 8;
    my $has_alpha = 0;
    my $trans_color = undef;
    while (!defined($self->{data})) {
        if (length($$data_ref) < $pos + 8) {
            print STDERR "${path}PNG data seems to be truncated\n";
            return undef;
        }
        my ($chunklen, $chunktype) = unpack("NA4", substr($$data_ref,$pos,8));
        $pos += 8;
        if (length($$data_ref) < $pos + ($chunklen + 4)) {
            print STDERR "${path}PNG $chunktype data seems to be truncated\n";
            return undef;
        }
        my $chunk = substr($$data_ref, $pos, $chunklen);
        $pos += $chunklen + 4;  # Ignore the CRC
        if ($chunktype eq "IHDR") {
            my ($width, $height, $bitdepth, $format, $compression,
                $filter, $interlace) = unpack("NNCCCCC", $chunk);
            if ($bitdepth != 8) {
                print STDERR "${path}Unsupported bit depth $bitdepth\n";
                return undef;
            }
            if ($format != 2 && $format != 6) {
                print STDERR "${path}Unsupported image format $format\n";
                return undef;
            }
            if ($compression != 0) {
                print STDERR "${path}Unsupported compression mode $compression\n";
                return undef;
            }
            if ($filter != 0) {
                print STDERR "${path}Unsupported filter type $filter\n";
                return undef;
            }
            if ($interlace != 0) {
                print STDERR "${path}Unsupported interlacing mode $interlace\n";
                return undef;
            }
            $self->{width} = $width;
            $self->{height} = $height;
            $has_alpha = ($format == 6);
        } elsif ($chunktype eq "tRNS") {
            if ($has_alpha) {
                print STDERR "${path}tRNS chunk not allowed in alpha images\n";
                return undef;
            }
            if (length($chunk) != 6) {
                print STDERR "${path}Invalid tRNS chunk size\n";
                return undef;
            }
            my (undef, $r, undef, $g, undef, $b) = unpack("C6", $chunk);
            $trans_color = 0xFF000000 | $r<<16 | $g<<8 | $b;
        } elsif ($chunktype eq "IDAT") {
            my $image = uncompress($chunk);
            if (!defined($image)) {
                print STDERR "${path}Error decompressing image data\n";
                return undef;
            }
            my $Bpp = ($has_alpha ? 4 : 3);
            $self->{data} = [];
            for (my $y = 0; $y < $self->{height}; $y++) {
                my $filter = unpack("C", substr($image, 0, 1));
                my @rowdata = unpack("C*", substr($image, 1,
                                                  $self->{width} * $Bpp));
                $image = substr($image, 1 + ($self->{width} * $Bpp));
                my @row = ();
                for (my $x = 0; $x < $self->{width}; $x++) {
                    my $pixel = $has_alpha ? 0 : 0xFF000000;
                    for (my $component = 0; $component < $Bpp; $component++) {
                        my $shift = ((2 - $component) & 3) * 8;
                        my $left = ($x==0 ? 0 : $row[$x-1]>>$shift & 0xFF);
                        my $up = ($y==0 ? 0 : $self->{data}[$y-1][$x]>>$shift & 0xFF);
                        my $upleft = ($y==0 ? 0 : $x==0 ? 0 : $self->{data}[$y-1][$x-1]>>$shift & 0xFF);
                        my $this = $rowdata[$x*$Bpp + $component];
                        if ($filter == 1) {
                            $this = ($this + $left) & 0xFF;
                        } elsif ($filter == 2) {
                            $this = ($this + $up) & 0xFF;
                        } elsif ($filter == 3) {
                            $this = ($this + int(($left+$up)/2)) & 0xFF;
                        } elsif ($filter == 4) {
                            my $p = $left + $up - $upleft;
                            my $pa = abs($p - $left);
                            my $pb = abs($p - $up);
                            my $pc = abs($p - $upleft);
                            my $pred;
                            if ($pa <= $pb && $pa <= $pc) {
                                $pred = $left;
                            } elsif ($pb <= $pc) {
                                $pred = $up;
                            } else {
                                $pred = $upleft;
                            }
                            $this = ($this + $pred) & 0xFF;
                        }
                        $pixel |= $this << $shift;
                    }
                    if (defined($trans_color) && $pixel == $trans_color) {
                        $pixel = 0;
                    }
                    push @row, $pixel;
                }
                push @{$self->{data}}, [@row];
            }
        }
    }

    return $self;
}

###########################################################################

# width, height:  Return the width or height of the image.
#
# [Parameters]
#     None
# [Return value]
#     Width or height of the stored image, in pixels

sub width()
{
    my ($self) = @_;
    return $self->{width};
}

sub height()
{
    my ($self) = @_;
    return $self->{height};
}

###########################################################################

# data:  Return the image data.  The data is stored in a two-dimensional
# array, with [0][0] containing the upper-left pixel of the image.  Each
# element of the array is a 32-bit pixel value in 0xAARRGGBB format; if the
# PNG data did not contain any transparency information, every pixel will
# have an alpha value of 0xFF (opaque).
#
# The returned data is a copy of the image data stored in the object, and
# may be modified freely.
#
# [Parameters]
#     None
# [Return value]
#     Reference to image data

sub data()
{
    my ($self) = @_;
    return [@{$self->{data}}];
}

###########################################################################
###################### Instance field documentation #######################
###########################################################################

# width, height
#     Size of the image, in pixels.
#
# data
#     Image data, organized as a two-dimensional array with [0][0] at the
#     top-left of the image.

###########################################################################

1;

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
