#!/usr/bin/perl
#
# Proof of concept for generating a BDMV directory structure from one or
# more M2TS inputs.
#
# Copyright (c) 2010 Andrew Church
# See the file COPYING for distribution conditions.

use strict;
use XML::Simple;

BEGIN {
    my $script_dir = $0;
    $script_dir =~ s,/[^/]*$,, or $script_dir = ".";
    push @INC, $script_dir;
}
use BDDisc;
use BDMenu;

###########################################################################

# Avoid wide character warnings
binmode STDERR, ":utf8";

my @clips = ();
my @clipinfo = ();
my $auto_chapter_freq = undef;
my @exclude_pids = ();
my $loop_at_end = 0;
my $add_menu = 0;
my $outdir = undef;
my $pana_mode = 0;
my $split_size = undef;
my $xml_input = undef;

for (my $i = 0; $i < @ARGV; $i++) {
    if ($ARGV[$i] =~ /^-(.*)/) {
        my $option = $1;
        if ($option eq "") {  # i.e. just "-"
            push @clips, @ARGV[$i+1..$#ARGV];
            last;
        } elsif ($option =~ s/^a//) {
            my $freqstr;
            if ($option ne "") {
                $freqstr = $option;
            } else {
                $freqstr = $ARGV[++$i];
            }
            $auto_chapter_freq = &parse_time($freqstr);
            if (!defined($auto_chapter_freq) || $auto_chapter_freq <= 0) {
                die "Bad time parameter to option -a: $freqstr\n";
            }
        } elsif ($option =~ s/^e//) {
            my $pidstr;
            if ($option ne "") {
                $pidstr = $option;
            } else {
                $pidstr = $ARGV[++$i];
            }
            my $pid;
            if ($pidstr =~ /^0x/i) {
                $pid = hex($pidstr);
            } else {
                $pid = 0 + $pidstr;
            }
            die "Bad PID parameter to option -e: $pidstr\n"
                if $pid < 1 || $pid >= 0x2000;
            push @exclude_pids, $pid;
        } elsif ($option =~ /^-?h/) {
            &usage;
        } elsif ($option eq "l") {
            $loop_at_end = 1;
        } elsif ($option eq "m") {
            $add_menu = 1;
        } elsif ($option =~ s/^o//) {
            die "Only one -o option may be given\n" if defined($outdir);
            if ($option ne "") {
                $outdir = $option;
            } else {
                $outdir = $ARGV[++$i];
            }
        } elsif ($option eq "p") {
            $pana_mode = 1;
        } elsif ($option =~ s/^s//) {
            if ($option ne "") {
                $split_size = $option;
            } else {
                $split_size = $ARGV[++$i];
            }
            if ($split_size =~ /^([0-9]+|[0-9]*\.[0-9]+)([kMG]B?)?$/i) {
                $split_size = $1;
                my $unit = lc($2);
                $split_size *= 1024           if $unit eq "k";
                $split_size *= 1000           if $unit eq "kb";
                $split_size *= 1024*1024      if $unit eq "m";
                $split_size *= 1000*1000      if $unit eq "mb";
                $split_size *= 1024*1024*1024 if $unit eq "g";
                $split_size *= 1000*1000*1000 if $unit eq "gb";
                $split_size = int($split_size);
            } else {
                die "Invalid split size: $split_size\n";
            }
        } elsif ($option =~ s/^x//) {
            die "Only one -x option may be given\n" if defined($xml_input);
            if ($option ne "") {
                $xml_input = $option;
            } else {
                $xml_input = $ARGV[++$i];
            }
        } else {
            print STDERR "Unrecognized option -$option\n";
            &usage;
        }
    } else {
	push @clips, $ARGV[$i];
    }
}

&usage if !@clips && !defined($xml_input);
&usage if defined($xml_input) && (@clips || @exclude_pids || $loop_at_end || $add_menu);

########

die "mkdir($outdir): $!\n" if ! -d $outdir && !mkdir($outdir);

my $bd = BDDisc->create($outdir, $pana_mode, $pana_mode, $split_size);

if ($xml_input) {

    &process_xml($bd, $xml_input);

} else {

    for (my $i = 0; $i < @clips; $i++) {
        my $post_action;
        if ($add_menu) {
            $post_action = "menu";
        } elsif ($i < @clips-1) {
            $post_action = "title=" . ($i+2);
        } elsif ($loop_at_end) {
            $post_action = "title=1";
        } else {
            $post_action = "standby";
        }
        $bd->add_stream($clips[$i], $auto_chapter_freq, $post_action,
                        \@exclude_pids);
    }

    if ($add_menu) {
        my $menu = BDSimpleMenu->create();
        my $index = 1;
        foreach my $clip (@clips) {
            my $title = $clip;
            $title =~ s|^.*/||;
            $title =~ s|\.m2?ts$||i;
            $menu->add_title($index, $title);
            $index++;
        }
        $bd->add_menu($menu);
    }

}

$bd->finalize();

########

exit 0;

###########################################################################

# usage:  Print a usage statement and exit with failure.
#
# [Parameters]
#     None
# [Return value]
#     None

sub usage()
{
    die "Usage: $0 [-a [[HH:]MM:]SS] [-e pid [-e pid...]] [-l] [-m] \\\n" .
        "          [-p] [-s split-size] -o outdir clip.m2ts...\n" .
        "   or: $0 [-p] [-s split-size] -o outdir -x control.xml\n" .
        "    -a: Add chapters every HH:MM:SS from the start of each stream\n" .
        "    -e: Exclude specified PIDs from output (decimal or 0xHEX)\n" .
        "    -l: Loop at end of last title instead of going to standby mode\n".
        "    -m: Add a simple menu listing the titles on the disc\n" .
        "    -o: Specify base directory for output files (REQUIRED)\n" .
        "    -p: Enable Panasonic Diga SDHC mode (BROKEN)\n" .
        "    -s: Split streams into files no larger than split-size bytes\n" .
        "           (128M, 3.5G, etc.)  Suffixes: k = 1024, kB = 1000,\n" .
        "           likewise for M, MB, G, GB\n" .
        "    -x: Read the given XML control file for options and clips\n" .
        "           (see documentation)\n";
}

###########################################################################

# process_xml:  Process an XML control file to create a BD file structure.
#
# [Parameters]
#       $bd: BDDisc object
#     $path: XML control file pathname
# [Return value]
#     None

sub process_xml($$)
{
    my ($bd, $path) = @_;

    my $xml = new XML::Simple;
    my $data = $xml->XMLin($path);
    die "Failed to read or parse $path\n" if !$data;
    die "No video titles specified in $path\n" if !$data->{title};

    my $has_menu = defined($data->{menu}) || defined($data->{simplemenu});

    my $default_post = $data->{defaults}{titlepost}
        || ($has_menu ? "menu" : "next");
    if ($default_post ne "standby"
     && $default_post ne "repeat"
     && $default_post ne "menu"
     && $default_post ne "next"
     && $default_post ne "next-loop"
     && $default_post ne "next-menu"
    ) {
        die "Invalid default title-post action \"$default_post\"\n";
    }
    if (!$has_menu && $default_post =~ /menu/) {
        die "Default title-post action \"$default_post\" requires a menu\n";
    }

    my $text_defaults = {
        align => 'left',
        font => 'Sans',
        size => 50,
        color => 'FFFFFFFF',
    };
    if ($data->{defaults}{textalign}) {
        if ($data->{defaults}{textalign} ne 'left'
         && $data->{defaults}{textalign} ne 'center'
         && $data->{defaults}{textalign} ne 'right'
        ) {
            die "Invalid default text alignment \"$data->{defaults}{textalign}\"\n";
        }
        $text_defaults->{align} = $data->{defaults}{textalign};
    }
    if ($data->{defaults}{textfont}) {
        $text_defaults->{font} = $data->{defaults}{textfont};
    }
    if ($data->{defaults}{textsize}) {
        if ($data->{defaults}{textsize} !~ /^\s*[1-9][0-9]*\s*$/) {
            die "Invalid default font size \"$data->{defaults}{textsize}\"\n";
        }
        $text_defaults->{size} = 0 + $data->{defaults}{textsize};
    }
    if ($data->{defaults}{textcolor}) {
        if ($data->{defaults}{textcolor} !~ /^\s*[0-9A-Fa-f]{8}\s*$/) {
            die "Invalid default font color \"$data->{defaults}{textcolor}\"\n";
        }
        $text_defaults->{color} = $data->{defaults}{textcolor};
    }

    my @titles;
    if (ref($data->{title}) eq "ARRAY") {
        @titles = @{$data->{title}};
    } else {
        @titles = ($data->{title});
    }
    my $titlenum = 0;
    foreach my $title (@titles) {
        $titlenum++;
        my $this_default_post;
        if ($default_post eq "repeat") {
            $this_default_post = "title=$titlenum";
        } elsif ($default_post =~ /^next/) {
            if ($titlenum < scalar(@titles)) {
                $this_default_post = "title=" . ($titlenum+1);
            } elsif ($default_post eq "next-loop") {
                $this_default_post = "title=1";
            } elsif ($default_post eq "next-menu") {
                $this_default_post = "menu";
            } else {
                $this_default_post = "standby";
            }
        } else {
            $this_default_post = $default_post;
        }
        &process_xml_title($bd, $titlenum, $title, $this_default_post);
    }

    if (defined($data->{menu})) {
        my $menu = BDMenu->create();
        &process_xml_menu($menu, $data->{menu}, $text_defaults);
        $bd->add_menu($menu);
    } elsif (defined($data->{simplemenu})) {
        my $menu = BDSimpleMenu->create();
        my $titlenum = 1;
        foreach my $title (@titles) {
            $menu->add_title($titlenum, $title->{name});
            $titlenum++;
        }
        $bd->add_menu($menu);
    }
}

######################################

# process_xml_title:  Process a <title> tag in an XML control file.
#
# [Parameters]
#               $bd: BDDisc object
#         $titlenum: Title number (1...)
#            $title: <title> tag data
#     $default_post: Default post-title action ("standby", "menu", "title=...")
# [Return value]
#     None
# [Side effects]
#     Sets $title->{name} to a name based on the source filename if it is
#     not already set.

sub process_xml_title($$$$)
{
    my ($bd, $titlenum, $title, $default_post) = @_;

    die "\"source\" attribute missing from <title> in title $titlenum\n"
        if !$title->{source};
    if (!defined($title->{name})) {
        $title->{name} = $title->{source};
        $title->{name} =~ s|^.*/||;
        $title->{name} =~ s|\.m2?ts$||i;
    }

    my $chapters = undef;
    if ($title->{chapters}) {
        $chapters = [];
        foreach my $time (split(/\s*,\s*/, $title->{chapters})) {
            my $sec = parse_time($time);
            if (!defined($sec)) {
                die "Bad value $time for <chapters> in title $titlenum ($title->{source})\n";
            } elsif (!@$chapters && $sec == 0) {
                # Ignore an initial zero
            } elsif (@$chapters && $sec <= $$chapters[@$chapters-1]) {
                die "Chapter time $time is not greater than previous time in title $titlenum ($title->{source})\n";
            } else {
                push @$chapters, $sec;
            }
        }
    } elsif ($title->{autochapters}) {
        $chapters = &parse_time($title->{autochapters})
            or die "Bad value $title->{autochapters} for <autochapters> in title $titlenum ($title->{source})\n";
    }

    my $post_action = $title->{post} || $default_post;

    my @exclude_pids = ();
    if (defined($title->{exclude})) {
        if (ref($title->{exclude}) eq "ARRAY") {
            @exclude_pids = @{$title->{exclude}};
        } else {
            @exclude_pids = ($title->{exclude});
        }
        map {$_ = $_->{pid} or die "\"pid\" attribute missing from <exclude> in title $titlenum ($title->{source})\n";} @exclude_pids;
        map {$_ = hex($_) if $_ =~ s/^0x//i;} @exclude_pids;
    }

    $bd->add_stream($title->{source}, $chapters, $post_action, \@exclude_pids);
}

######################################

# process_xml_menu:  Process a <menu> tag in an XML control file.
#
# [Parameters]
#              $menu: BDMenu object
#              $data: <menu> tag data
#     $text_defaults: Default text attributes
# [Return value]
#     None

sub process_xml_menu($$)
{
    my ($menu, $data, $text_defaults) = @_;

    my $common_background = {color => '000000'};
    if (defined($data->{background})) {
        if (ref($data->{background}) eq "ARRAY") {
            die "Only one <background> tag may be present in a <menu> tag\n";
        } elsif (defined($data->{background}{color})
                 && $data->{background}{color} !~ /^\s*[0-9A-Fa-f]{6}\s*$/) {
            die "Invalid background color \"$data->{background}{color}\"\n";
        } elsif (defined($data->{background}{source})
                 && !(-r $data->{background}{source})) {
            die "Background image \"$data->{background}{source}\" not readable\n";
        } elsif (defined($data->{background}{resize})
                 && $data->{background}{resize} !~ /^\s*(linear|cubic|sinc)(\s*,\s*aspect)?\s*$/) {
            die "Invalid background resize operation \"$data->{background}{resize}\"\n";
        } else {
            $common_background = $data->{background};
            $common_background->{color} ||= '000000';
        }
    }
    my @common_text = ();
    if (defined($data->{text})) {
        if (ref($data->{text}) eq "ARRAY") {
            @common_text = @{$data->{text}};
        } else {
            @common_text = ($data->{text});
        }
    }
    my @common_links = ();
    if (defined($data->{link})) {
        if (ref($data->{link}) eq "ARRAY") {
            @common_links = @{$data->{link}};
        } else {
            @common_links = ($data->{link});
        }
    }

    my $image_cache = {};

    my @pages;
    if (ref($data->{page}) eq "ARRAY") {
        @pages = @{$data->{page}};
    } else {
        @pages = ($data->{page});
    }

    my $pagenum = 0;
    foreach my $page (@pages) {
        $pagenum++;

        my $background = $common_background;
        if (defined($page->{background})) {
            if (ref($page->{background}) eq "ARRAY") {
                die "Only one <background> tag may be present in a <page> tag\n";
            } elsif (defined($page->{background}{color})
                     && $page->{background}{color} !~ /^\s*[0-9A-Fa-f]{6}\s*$/) {
                die "Invalid background color \"$page->{background}{color}\"\n";
            } elsif (defined($page->{background}{source})
                     && !(-r $page->{background}{source})) {
                die "Background image \"$page->{background}{source}\" not readable\n";
            } elsif (defined($page->{background}{resize})
                     && $page->{background}{resize} !~ /^\s*(linear|cubic|sinc)(\s*,\s*aspect)?\s*$/) {
                die "Invalid background resize operation \"$page->{background}{resize}\"\n";
            } else {
                $background = $page->{background};
                $background->{color} ||= $common_background->{color};
            }
        }
        $menu->set_background($pagenum-1,
                              hex($background->{color} || '000000'),
                              $background->{source}, $background->{resize});

        my @text = ();
        if (defined($page->{text})) {
            if (ref($page->{text}) eq "ARRAY") {
                @text = @{$page->{text}};
            } else {
                @text = ($page->{text});
            }
        }
        foreach my $text (@common_text, @text) {
            $text->{align} ||= $text_defaults->{align};
            $text->{font}  ||= $text_defaults->{font};
            $text->{size}  ||= $text_defaults->{size};
            $text->{color} ||= $text_defaults->{color};
            &process_xml_menu_text($menu, $pagenum, $text);
        }

        my @links = ();
        if (defined($page->{link})) {
            if (ref($page->{link}) eq "ARRAY") {
                @links = @{$page->{link}};
            } else {
                @links = ($page->{link});
            }
        }
        foreach my $link (@common_links, @links) {
            &process_xml_menu_link($menu, $pagenum, $link, $image_cache);
        }

    }
}

######################################

# process_xml_menu_text:  Process a menu <text> tag in an XML control file.
#
# [Parameters]
#        $menu: BDMenu object
#     $pagenum: Menu page number (1...)
#        $data: <text> tag data
# [Return value]
#     None

sub process_xml_menu_text($$$)
{
    my ($menu, $pagenum, $data) = @_;

    die "Text tag missing coordinates for text \"$data->{content}\"\n"
        if !defined($data->{x}) || !defined($data->{y});
    # Everything else is assumed to be set by the caller.

    my $align;
    if ($data->{align} eq 'left') {
        $align = -1;
    } elsif ($data->{align} eq 'center') {
        $align =  0;
    } elsif ($data->{align} eq 'right') {
        $align = +1;
    } else {
        die "Invalid alignment \"$data->{align}\" for text \"$data->{content}\"\n";
    }

    my $color;
    die "Invalid color \"$data->{color}\" for text \"$data->{content}\"\n"
        if $data->{color} !~ /^\s*([0-9A-Fa-f]{8})\s*$/;
    $color = hex($1);

    $menu->add_text($pagenum-1, $data->{x}, $data->{y}, $align,
                    $data->{font}, $data->{size}, $color, $data->{content});
}

######################################

# process_xml_menu_link:  Process a menu <link> tag in an XML control file.
#
# [Parameters]
#            $menu: BDMenu object
#         $pagenum: Menu page number (1...)
#            $data: <text> tag data
#     $image_cache: Image cache (initialize to an empty hash reference)
# [Return value]
#     None

sub process_xml_menu_link($$$$)
{
    my ($menu, $pagenum, $data, $image_cache) = @_;

    $data->{index} = 0 + $data->{index};
    die "Link tag missing coordinates for index $data->{index}\n"
        if (!defined($data->{x}) || !defined($data->{y}))
        && ($data->{normal} || $data->{hilite} || $data->{active});

    my ($normal, $hilite, $active);
    sub load_image($$) {
        my ($path) = @_;
        if (!defined($image_cache->{$path})) {
            my $png = PNGReader->read($path)
                or die "Failed to load image \"$path\" for link index $data->{index}\n";
            $image_cache->{$path} = $menu->png_to_image($png);
        }
        return $image_cache->{$path};
    }
    if ($data->{normal}) {
        $normal = &load_image($data->{normal});
    }
    if ($data->{hilite}) {
        $hilite = &load_image($data->{hilite});
    }
    if ($data->{active}) {
        $active = &load_image($data->{active});
    }

    my @link_params = (
        $pagenum-1, $data->{index},
        $data->{x}, $data->{y}, $normal, $hilite, $active,
        $data->{up}, $data->{down}, $data->{left}, $data->{right},
        defined($data->{auto})
    );
    if ($data->{content} =~ /^\s*title=0*([1-9][0-9]*)(?:\.0*([1-9][0-9]*))?$/) {
        my $target = $1;
        my $chapter = $2;
        $menu->add_link_to_title(@link_params, $target, $chapter);
    } elsif ($data->{content} =~ /^\s*page=0*([1-9][0-9]*)\s*$/) {
        my $target = $1;
        $menu->add_link_to_page(@link_params, $target-1);
    } else {
        die "Invalid link target \"$data->{content}\" for index $data->{index}\n";
    }
}

###########################################################################

# parse_time:  Parse a time value in the form [[HH:]MM:]SS[.sss] and return
# the number of seconds it specifies.
#
# [Parameters]
#     $time: Time string
# [Return value]
#     Equivalent time in seconds, or undef on an invalid format

sub parse_time($)
{
    my ($time) = @_;
    if ($time !~ /^(?:([0-9]+:)?([0-9]+:))?([0-9]+(?:\.[0-9]+)?)$/) {
        return undef;
    }
    return $1*3600 + $2*60 + $3;
}

###########################################################################

# Local variables:
#   indent-tabs-mode: nil
# End:
#
# vim: expandtab shiftwidth=4:
